package edu.cosy.netgen.ann;

import java.io.*;
import java.util.*;
import java.beans.*;
import java.lang.reflect.Method;

/**
 * Provides a bean for Neural Networks.
 */
public class ANN implements Serializable {
    
    // name of the default activation function
    private final static String DEFAULT_ACT_FUNCTION = "ActFunctionSigmoid";
    private final static String DEFAULT_FIT_FUNCTION = "FitnessFunctionMSE";
    private final static String DEFAULT_TRAIN_FUNCTION = "Rprop";

    // contains the net structure
    private NeuralNet nn;
    
    // current fitness functions
    private FitnessFunction fitnessFunction;
    
    // current training function
    private Training trainingFunction;

    // current regularization function
    private Regularization regFunction;
    
    // current network type
    private int netType;

    // min. cycles a net must trained
    private int maxTrainingCycles = 1000;

    // shows how many cycles a net has been trained
    private int trainedCycles = -1;

    // when to stop training
    private double stopError = 0.00001;

    // time of training
    private long trainingTime = -1;
    
    // description of the bean
    private String description = null;
        
    // pattern set for training
    private transient PatternSet trainingPatternSet;
    
    // pattern set for validation
    private transient PatternSet validationPatternSet;
    
    // pattern set for calculation
    private transient PatternSet calculationPatternSet;
                                                               
    // pool for the fitness functions
    private Hashtable fitnessFunctions;
    
    // pool for the activation functions
    private Hashtable activationFunctions;
    
    // pool for the regularization functions
    private Hashtable regFunctions;
    
    // pool for training functions   
    private Hashtable trainingFunctions;
    
    // listener for training events
    private transient Vector trainingEventTargets;
    
    // listener for progress events
    private transient Vector progressEventTargets;
    
    // property change listeners
    private transient PropertyChangeSupport changes;
    
    /**
     * Create a new bean instance.
     */
    public ANN() {
        super();
        
        // demo initialization of description
        setDescription( "unknown" );
        
        // demo initialization of network structure
        setNetType( NeuralNet.FEEDFORWARD );
                
        // init function pools
        fitnessFunctions = new Hashtable();
        fitnessFunctions.put( "FitnessFunctionClassification", FitnessFunctionClassification.class ); 
        fitnessFunctions.put( "FitnessFunctionSSE", FitnessFunctionSSE.class ); 
        fitnessFunctions.put( "FitnessFunctionMSE", FitnessFunctionMSE.class ); 
        
        activationFunctions = new Hashtable();
        activationFunctions.put( "ActFunctionLinear", ActFunctionLinear.class );
        activationFunctions.put( "ActFunctionSigmoid", ActFunctionSigmoid.class );
        activationFunctions.put( "ActFunctionTanH", ActFunctionTanH.class );
        
        regFunctions = new Hashtable();
        regFunctions.put( "RegularizationConnection", RegularizationConnection.class );
        
        trainingFunctions = new Hashtable();
        trainingFunctions.put( "BP", BP.class );
        trainingFunctions.put( "Rprop", Rprop.class );

        // demo initialization of training function    
        setTrainingFunctionByName( DEFAULT_TRAIN_FUNCTION );
        
        // demo initialization of pattern sets
        setTrainingPatternSet( new PatternSet() );
        setValidationPatternSet( new PatternSet() );
        setCalculationPatternSet( new PatternSet() );

        // demo initialization of fitness function
        setFitnessFunctionByName( DEFAULT_FIT_FUNCTION );

        changes = new PropertyChangeSupport( this );
    }
        
    /**
     * Set the maximum number of cycles to be trained. 
     * If the value is 0, then the training cycles are not limited.
     * A smaller value is set to 0.
     * <P>
     * @param value maximum number of training cycles
     */
    public synchronized void setMaxTrainingCycles( int value ) {
        int oldValue = maxTrainingCycles;
        maxTrainingCycles = value > 0 ? value : 0;
        if ( changes != null )
            changes.firePropertyChange( "maxTrainingCycles", oldValue, maxTrainingCycles );        
    }
        
    /**
     * Get the maximum number of training cycles.
     * <P>
     * @return maximum number of training cycles
     */
    public int getMaxTrainingCycles() {
        return maxTrainingCycles;
    }
    
    /**
     * Set the description of the bean.
     * <P>
     * @param value new description
     */ 
    public synchronized void setDescription( String value ) {
        String oldValue = description;
        description = value;
        if ( changes != null )
            changes.firePropertyChange( "description", oldValue, value );
    }
    
    /**
     * Get the description of the bean.
     * <P>
     * @return description of the bean
     */
    public String getDescription() {
        return description;
    }
    
    /**
     * Set the stop error for training. Value 0.0 indicates that the training is not limited.
     * If the value is smaller than 0.0, it is set to 0.0.
     * <P>
     * @param value stop error for training
     */
    public synchronized void setStopError( double value ) {
        Double oldValue = new Double( value );
        stopError = value > 0.0 ? value : 0.0;
        if ( changes != null )
            changes.firePropertyChange( "stopError", oldValue, new Double( stopError ) );
    }
    
    /**
     * Get the stop error for training.
     * <P>
     * @return stop error for training
     */
    public double getStopError() {
        return stopError;
    }

    /**
     * Set the Markers for neurons. If a marker for one neuron is set, 
     * then it is not included in the training algorithm. 
     * The number of markers and neurons must be equal.
     * <P>
     * @param value array of marker values
     */
    public synchronized void setNeuronMarkers( boolean[] values ) throws java.lang.IllegalArgumentException {
        if ( getConMatrix() != null && values != null ) {
            boolean[] oldValues = getConMatrix().getMarkers();
            getConMatrix().setMarkers( values );
            nn.setMarkersModus( true );
            if ( changes != null )
                changes.firePropertyChange( "neuronMarkers", oldValues, values );            
        }
    }
    
    /**
     * Get the neuron markers of the net.
     * <P>
     * @return array of markers
     */
    public boolean[] getNeuronMarkers() {
        boolean[] retVal = null;
        if ( getConMatrix() != null )
            retVal = getConMatrix().getMarkers();
        return retVal;
    }

    /**
     * Get the possible regularization function names.
     * <P>
     * @return array of function names
     */
    public String[] getRegFunctionNames() {
        String[] retVal = new String[ regFunctions.size() ];
        Enumeration keys = regFunctions.keys();
        int i = 0;
        while ( keys.hasMoreElements() ) {    
            retVal[ i ] = ( String ) keys.nextElement();
            i++;
        }
        return retVal;
    }

    /**
     * Set the function for regularization by name. If there is no function with this name,
     * it is set to null.
     * <P>
     * @param value regularization function name
     */ 
    public synchronized void setRegFunctionByName( String value ) {
        Object oldValue = regFunction;
        if ( value != null && regFunctions.containsKey( value ) ) {
            try {
                regFunction = ( Regularization ) ( ( Class ) regFunctions.get( value ) ).newInstance();
            } catch ( Exception e ) {
                regFunction = null;
            }
        }
        else
            regFunction = null;
        if ( changes != null )
            changes.firePropertyChange( "regFunction", oldValue, regFunction );
    }
    
    /**
     * Get the name of the current function for regularization.
     * <P>
     * @return name of the regularization function
     */
    public String getRegFunctionByName() {
        String retVal = null;
        if ( regFunction != null ) {
            Enumeration keys = regFunctions.keys();
            String tmpKey = null;
            while ( keys.hasMoreElements() ) {
                tmpKey = ( String ) keys.nextElement();
                if ( ( ( Class ) regFunctions.get( tmpKey ) ).isInstance( regFunction ) ) {
                    retVal = tmpKey;
                    break;
                }
            }
        }
        return retVal;
    }
    
    /**
     * Set the function for regularization. If there is no function with this type,
     * it is set to null.
     * <P>
     * @param value regularization function
     */ 
    public synchronized void setRegFunction( Regularization value ) {
        Object oldValue = regFunction;
        if ( value != null && regFunctions.contains( value ) ) {
            regFunction = value;
        }
        else
            regFunction = null;
        if ( changes != null )
            changes.firePropertyChange( "regFunction", oldValue, regFunction );
    }
    
    /**
     * Get the current function for regularization.
     * <P>
     * @return the current regularization function
     */
    public Regularization getRegFunction() {
        return regFunction;
    }
    
/*****************************************************************************************************************************
 **
 ** Training Functions
 **
 ****************************************************************************************************************************/
  
    /**
     * Get the possible training function names.
     * <P>
     * @return array of function names
     */
    public String[] getTrainingFunctionNames() {
        String[] retVal = new String[ trainingFunctions.size() ];
        Enumeration keys = trainingFunctions.keys();
        int i = 0;
        while ( keys.hasMoreElements() ) {    
            retVal[ i ] = ( String ) keys.nextElement();
            i++;
        }
        return retVal;
    }

    /**
     * Set the function for training by name. If there is no function with this name,
     * java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param value training function name
     */ 
    public synchronized void setTrainingFunctionByName( String value ) throws java.lang.IllegalArgumentException {
        boolean isError = true;
        if ( trainingFunctions.containsKey( value ) ) {
            Object oldValue = trainingFunction;
            try {
                trainingFunction = ( Training ) ( ( Class ) trainingFunctions.get( value ) ).newInstance();
            } catch ( Exception e ) {
            }
            if ( changes != null )
                changes.firePropertyChange( "trainingFunction", oldValue, trainingFunction );
            isError = false;
        }
        if ( isError )
            throw new IllegalArgumentException();
    }
    
    /**
     * Get the name of the current function for training.
     * <P>
     * @return name of the training function
     */
    public String getTrainingFunctionByName() {
        String retVal = null;
        String tmp = null;
        Enumeration keys = trainingFunctions.keys();
        while ( keys.hasMoreElements() ) {              
            tmp = ( String ) keys.nextElement();
            if ( ( ( Class ) trainingFunctions.get( tmp ) ).isInstance( trainingFunction ) )
                retVal = tmp;
        }
        return retVal;
    }
    
    /**
     * Set the training parameters for the current training function. 
     * If the parameters can not be set, a java.lang.IllegalArgumentException is thrown.
     * <P>
     * @params value array of training paramters
     */
    public void setTrainingFunctionParams( TrainingParam[] values ) throws IllegalArgumentException {
        if ( trainingFunction != null ) {
            trainingFunction.setParams( values );
        }
    }

    /**
     * Get the parameters for the current training function.
     * <P>
     * @return array of training parameters
     */
    public TrainingParam[] getTrainingFunctionParams() {
        if ( trainingFunction != null )
            return trainingFunction.getParams();
        else
            return null;
    }
    
    /**
     * Set the training function for this bean. An java.lang.IllegalArgumentException is thrown, 
     * if it is an unknown function.
     * <P>
     * @param value training function
     */
    public void setTrainingFunction( Training value ) throws java.lang.IllegalArgumentException {
        if ( value != null && trainingFunctions.contains( value.getClass() ) ) {   
            Object oldValue = trainingFunction;         
            trainingFunction = value;
            if ( changes != null )
                changes.firePropertyChange( "trainingFunction", oldValue, trainingFunction );
        }
        else
            throw new IllegalArgumentException();
    }
    
    /**
     * Get the current training function.
     * <P>
     * @return current training function
     */
    public Training getTrainingFunction() {
        return trainingFunction;
    }
    
/*****************************************************************************************************************************
 **
 ** Pattern Sets
 **
 ****************************************************************************************************************************/
 
    /**
     * Set pattern values for network training.
     * <P>
     * @param value pattern set for training
     */   
    public synchronized void setTrainingPatternSet( PatternSet value ) {
        Object oldValue = trainingPatternSet;
        trainingPatternSet = value;
        if ( changes != null )
            changes.firePropertyChange( "trainingPatternSet", oldValue, trainingPatternSet );
    }

    /**
     * Get training pattern set.
     * <P>
     * @return pattern set for training
     */
    public PatternSet getTrainingPatternSet() {
           return trainingPatternSet;
    }
    
    /**
     * Set markers for the training patterns.
     * <P>
     * @param values array of markers
     */
    public synchronized void setTrainingPatternMarkers( boolean[] values ) {
        boolean[] oldValue = trainingPatternSet.getMarkers();
        trainingPatternSet.setMarkers( values );
        if ( changes != null )
            changes.firePropertyChange( "trainingPatternMarker", oldValue, values );
    }

    /**
     * Get the markers of the training pattern set.
     * <P>
     * @return array of markers
     */
    public boolean[] getTrainingPatternMarkers() {
        return trainingPatternSet.getMarkers();
    }

    /**
     * Set pattern values for validation.
     * <P>
     * @param value pattern set for validation
     */   
    public synchronized void setValidationPatternSet( PatternSet value ) {
        Object oldValue = validationPatternSet;
        validationPatternSet = value;
        if ( fitnessFunction != null )         
            fitnessFunction.init( validationPatternSet, ( FeedForwardNet ) nn );
        if ( changes != null )
            changes.firePropertyChange( "validationPatternSet", oldValue, validationPatternSet );
    }

    /**
     * Get validation pattern set.
     * <P>
     * @return pattern set for validation
     */
    public PatternSet getValidationPatternSet() {
           return validationPatternSet;
    }

    /**
     * Set pattern values for calculation.
     * <P>
     * @param value pattern set for calculation
     */   
    public synchronized void setCalculationPatternSet( PatternSet value ) {
        Object oldValue = calculationPatternSet;
        calculationPatternSet = value;
        if ( changes != null )
            changes.firePropertyChange( "calculationPatternSet", oldValue, calculationPatternSet );
    }
    
    /**
     * Get calculation pattern set
     * <P>
     * @return pattern set for calculation
     */
    public PatternSet getCalculationPatternSet() {
        return calculationPatternSet;
    }
    
/*****************************************************************************************************************************
 **
 ** Activation Function
 **
 ****************************************************************************************************************************/

    /**
     * Get the possible activation function names.
     * <P>
     * @return array of function names
     */
    public String[] getActivationFunctionNames() {
        String[] retVal = new String[ activationFunctions.size() ];
        Enumeration keys = activationFunctions.keys();
        int i = 0;
        while ( keys.hasMoreElements() ) {    
            retVal[ i ] = ( String ) keys.nextElement();
            i++;
        }
        return retVal;
    }

    /**
     * Set the activation functions for neurons. If a function is not known,
     * the default activation function is set. If the length of array is not equal to the number 
     * of neurons, then java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param values array of activation function names
     */
    public void setActivationFunctionsByNames( String[] values ) throws java.lang.IllegalArgumentException {
        boolean isError = true;
        if ( values != null ) {
            ActFunction[] functions = null;
            ActFunction[] oldValues = nn.getActFunctions();
            try {
                Object tmp = null;
                functions = new ActFunction[ values.length ];
                for ( int i = 0; i < functions.length; i++ ) {
                    if ( values[ i ] != null )
                        tmp = activationFunctions.get( values[ i ] );
                    else
                        tmp = activationFunctions.get( DEFAULT_ACT_FUNCTION );
                    functions[ i ] = ( ActFunction ) ( ( Class ) tmp ).newInstance();
                }
                nn.setActFunctions( functions );
                if ( changes != null )
                    changes.firePropertyChange( "actFunctions", oldValues, functions );
                isError = false;
            } catch ( Exception e ) {
            }  
        }
        if ( isError )
            throw new IllegalArgumentException();                 
    }
    
    /**
     * Get the names of the activation functions from the network.
     * <P>
     * @return array of function names
     */
    public String[] getActivationFunctionsByNames() {
        String[] retVal = new String[ nn.getActFunctions().length ];
        for ( int i = 0; i < retVal.length; i++ )
            retVal[ i ] = getActFunctionName( nn.getActFunctions()[ i ] );
        return retVal;
    }
    
    /**
     * Set the activation functions for neurons. If a function is not known,
     * the default activation function is set. If the length of array is not equal to the number 
     * of neurons, then java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param values array of activation functions
     */
    public void setActivationFunctions( ActFunction[] values ) throws java.lang.IllegalArgumentException {
        boolean isError = true;
        if ( values != null ) {
                ActFunction[] oldValues = nn.getActFunctions();
                nn.setActFunctions( values );
                if ( changes != null )
                    changes.firePropertyChange( "actFunctions", oldValues, values );
                isError = false;
        }
        if ( isError )
            throw new IllegalArgumentException();                 
    }
    
    /**
     * Get the functions from the network.
     * <P>
     * @return array of functions
     */
    public ActFunction[] getActivationFunctions() {
        return nn.getActFunctions();
    }
        
/*****************************************************************************************************************************
 **
 ** Fitness Function
 **
 ****************************************************************************************************************************/

    /**
     * Get the possible fitness function names.
     * <P>
     * @return array of function names
     */
    public String[] getFitnessFunctionNames() {
        String[] retVal = new String[ fitnessFunctions.size() ];
        Enumeration keys = fitnessFunctions.keys();
        int i = 0;
        while ( keys.hasMoreElements() ) {    
            retVal[ i ] = ( String ) keys.nextElement();
            i++;
        }
        return retVal;
    }
    
    /**
     * Set the function for fitness calculation by name. If there is no function with this name,
     * java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param value fitness function name
     */ 
    public synchronized void setFitnessFunctionByName( String value ) throws java.lang.IllegalArgumentException {
        boolean isError = true;
        FitnessFunction oldValue = fitnessFunction;
        try {
            if ( fitnessFunctions.containsKey( value ) ) 
                fitnessFunction = ( FitnessFunction ) (( Class ) fitnessFunctions.get( value )).newInstance();
        
            // set the new function if it's not null
            if ( fitnessFunction != null ) {
                fitnessFunction.init( validationPatternSet, (FeedForwardNet) nn );
                if ( changes != null )
                    changes.firePropertyChange( "fitnessFunction", oldValue, fitnessFunction );
                isError = false;      
            }
        } catch ( Exception e ) {    
            fitnessFunction = oldValue;
        }
        if ( isError )
            throw new IllegalArgumentException();
    }

    /**
     * Get the name of the current function for fitness calculation.
     * <P>
     * @return name of the fitness function
     */
    public String getFitnessFunctionByName() {
        String retVal = null;
        String tmp = null;
        Enumeration keys = fitnessFunctions.keys();
        while ( keys.hasMoreElements() ) {              
            tmp = ( String ) keys.nextElement();
            if ( ( ( Class ) fitnessFunctions.get( tmp ) ).isInstance( fitnessFunction ) )
                retVal = tmp;
        }
        return retVal;
    }

    /**
     * Set the fitness function for this bean. An java.lang.IllegalArgumentException is thrown, 
     * if it is an unknown function.
     * <P>
     * @param value fitness function
     */
    public synchronized void setFitnessFunction( FitnessFunction value ) throws java.lang.IllegalArgumentException {
        if ( value != null && fitnessFunctions.contains( value.getClass() ) ) {   
            Object oldValue = fitnessFunction;         
            fitnessFunction = value;
            fitnessFunction.init( validationPatternSet, (FeedForwardNet) nn );
            if ( changes != null )
                changes.firePropertyChange( "fitnessFunction", oldValue, fitnessFunction );
        }
        else
            throw new IllegalArgumentException();
    }
    
    /**
     * Get the current function for fitness calculation.
     * <P>
     * @return current fitness function
     */
    public FitnessFunction getFitnessFunction() {
        return fitnessFunction;
    }
    
    /**
     * Get the last training time for this Bean. The value is -1.0 there is no time value.
     * <P>
     * @return training time in seconds
     */
    public double getTrainingTime() {
        double tmp = -1;
        if ( trainingTime != -1 ) {
            tmp = ( ( double ) trainingTime ) / 1000;
        }
        return tmp;
    }        
    
    /**
     * Get the fitness of the Bean after training. -1.0 indicates that no fitness function is set.
     * If an regularization function is set, the value is regularized.
     * <P>
     * @return fitness value of the net
     */
    public double getFitness() {
        double retVal = -1;
        if ( fitnessFunction != null )
            // if a regularization function is here
            if ( regFunction != null )
                retVal = regFunction.getRegFitness( fitnessFunction.getFitness() );
            else
                retVal = fitnessFunction.getFitness();
        return retVal;
    }

    /**
     * Get the number of trained cycles. Default return value is -1.
     * <P>
     * @return trained cycles of the net
     */
    public int getTrainedCycles() {
        return trainedCycles;
    }

    /**
     * Set the number of input neurons.
     * <P>
     * @param value number of input neurons.
     */
    public synchronized void setIn( int value ) throws java.lang.IllegalArgumentException {
        if ( nn != null ) {
            if ( value >= 0 ) {
                int oldValue = nn.getIn();
                nn.setIn( value );     
                if ( changes != null )
                    changes.firePropertyChange( "in", oldValue, value );
            }
            else
                throw new java.lang.IllegalArgumentException();
        }
    }
    
    /**
     * Get the number of input neurons.
     * <P>
     * @return number of input neurons
     */
    public int getIn() {
        return nn.getIn();
    }
    
    /**
     * Set the number of hidden neurons.
     * <P>
     * @param value number of hidden neurons.
     */
    public synchronized void setHidden( int value ) throws java.lang.IllegalArgumentException {
        if ( nn != null ) {
            if ( value >= 0 ) {
                int oldValue = nn.getHidden();
                nn.setHidden( value );     
                if ( changes != null )
                    changes.firePropertyChange( "hidden", oldValue, value );
            }
            else
                throw new java.lang.IllegalArgumentException();
        }
    }
    
    /**
     * Get the number of hidden neurons.
     * <P>
     * @return number of hidden neurons
     */
    public int getHidden() {
        return nn.getHidden();
    }

    /**
     * Set the number of output neurons.
     * <P>
     * @param value number of output neurons.
     */
    public synchronized void setOut( int value ) throws java.lang.IllegalArgumentException {
        if ( nn != null ) {
            if ( value >= 0 ) {
                int oldValue = nn.getOut();
                nn.setOut( value );     
                if ( changes != null )
                    changes.firePropertyChange( "out", oldValue, value );
            }
            else
                throw new java.lang.IllegalArgumentException();
        }
    }
    
    /**
     * Get the number of output neurons.
     * <P>
     * @return number of output neurons
     */
    public int getOut() {
        return nn.getOut();
    }

    /**
     * Set the connection matrix for this bean. 
     * The length of the new connection matrix bust be equal to the old one, 
     * otherwise java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param value new connection matrix
     */
    public synchronized void setConMatrix( ConnectionMatrix value ) throws java.lang.IllegalArgumentException {
        if ( value != null ) {
            ConnectionMatrix oldValue = nn.getConMatrix();
            if ( trainingFunction != null ) {
                switch ( trainingFunction.getValidNetType() ) {
                    case NeuralNet.FEEDFORWARD: {
                        value.checkForward();                        
                        break;
                    }
                    case NeuralNet.RECURRENT: {                    
                        break;
                    }
                }
            }
            nn.setConMatrix( value );
            if ( changes != null )
                changes.firePropertyChange( "conMatrix", oldValue, value );            
        }
        else
            throw new IllegalArgumentException();
    }
    
    /**
     * Get a matrix of the connections of the bean.
     * <P>
     * @return connection matrix
     */
    public ConnectionMatrix getConMatrix() {
        return nn.getConMatrix();
    }

    /**
     * Set the connection style for the network. Possible values are:
     * <P>
     * ConnectionMatrix.CREATE_NONE:         no connections
     * ConnectionMatrix.CREATE_THREE_LAYER:  only connections between layers
     * ConnectionMatrix.CREATE_FULL:         all paossible connections
     * ConnectionMatrix.CREATE_RANDOM:       randomly connected
     * ConnectionMatrix.CREATE_BIAS_ONLY:    only connections for bias neurons
     * <P>
     * An unknown style throws a java.lang.IllegalArgumentException.
     * <P>
     * @param value connection style
     */
    public synchronized void setConnectionStyle( int value ) throws java.lang.IllegalArgumentException  {
        if ( nn != null && nn.conMatrix != null ) {
            int oldValue = nn.conMatrix.getConnectionStyle();
            nn.conMatrix.setConnectionStyle( value );
            if ( changes != null )
                changes.firePropertyChange( "connectionStyle", oldValue, value );            
        }
    }
    
    /**
     * Get the connection style of the network.
     * <P>
     * @return connection style
     */
    public int getConnectionStyle() {
        if ( nn != null && nn.conMatrix != null )
            return nn.conMatrix.getConnectionStyle();
        else
            return -1;
    }
    
    /**
     * Set the weights of the network.
     * The length of the new weight matrix bust be equal to the old one, 
     * otherwise java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param value new weight matrix
     */
    public synchronized void setWeightMatrix( WeightMatrix value ) throws java.lang.IllegalArgumentException {
        if ( value != null ) {
            WeightMatrix oldValue = nn.getWeightMatrix();
            nn.setWeightMatrix( value );
            if ( changes != null )
                changes.firePropertyChange( "weightMatrix", oldValue, value );            
        }
        else
            throw new IllegalArgumentException();
    }
    
    /**
     * Get the weights of the network.
     * <P>
     * @return weight matrix
     */
    public WeightMatrix getWeightMatrix() {
        return nn.getWeightMatrix();
    }
    
    /**
     * Set the network type. Currently, only one type is specified: NeuralNet.FEEDFORWARD.
     * A new network with this type is created.
     * If the type is unknown, then a java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param value type of the network
     */
    public synchronized void setNetType( int value ) throws java.lang.IllegalArgumentException {
        int oldValue = netType;
        switch ( value ) {
            case NeuralNet.FEEDFORWARD: {
                netType = NeuralNet.FEEDFORWARD;
                nn = new FeedForwardNet();
                if ( changes != null )
                    changes.firePropertyChange( "netType", oldValue, netType );
                break;
            }
            case NeuralNet.RECURRENT: {
                netType = NeuralNet.RECURRENT;
                if ( changes != null )
                    changes.firePropertyChange( "netType", oldValue, netType );
            }   
            default:
                throw new IllegalArgumentException();
        }
    }

    /**
     * Get the network type.
     * <P>
     * @return network type
     */
    public int getNetType() {
        return netType;
    }
        
    /**
     * Train the network.
     * <P>
     * @return true, if the training was successfull
     */
    public synchronized boolean trainNet() {
        boolean retVal = false;
        if ( nn != null && trainingFunction != null && fitnessFunction != null ) {
            switch ( trainingFunction.getValidNetType() ) {
                case NeuralNet.FEEDFORWARD: {
                    fitnessFunction.init( validationPatternSet, ( FeedForwardNet ) nn );
                    trainingFunction.init( ( FeedForwardNet ) nn, trainingPatternSet );
                    break;
                }
                case NeuralNet.RECURRENT: {                 
                    break;
                }
            }
            notifyTrainingEventTargets( TrainingEvent.TRAINING_INITIALIZED );
            long startTime = System.currentTimeMillis();
            notifyTrainingEventTargets( TrainingEvent.TRAINING_STARTED );
            trainedCycles = trainingFunction.train( this );
            notifyTrainingEventTargets( TrainingEvent.TRAINING_STOPED );
            trainingTime = System.currentTimeMillis() - startTime;
            retVal = true;
        }
        return retVal;
    }

    /**
     * Calculate the output of the network.
     * <P>
     * @return resultMatrix, each row stands for a pattern and the columns for the output neurons
     */
    public synchronized double[][] calculateNet() {
        double[][] retVal = null;
        if ( nn != null && calculationPatternSet != null ) {
            switch ( netType ) {
                case NeuralNet.FEEDFORWARD: {
                    FP fpCalc = new FP( ( FeedForwardNet ) nn, calculationPatternSet );
                    fpCalc.forwardStep();
                    retVal = fpCalc.getResultMatrix();
                    break;
                }
                case NeuralNet.RECURRENT: {
                    break;
                }
            }
        }
        return retVal;
    }
    
    /**
     * Add a listener for ProgressEvent.
     * <P>
     * @param l listener to add
     */
    public synchronized void addProgressListener( ProgressListener l ) {
        if ( progressEventTargets == null )
            progressEventTargets = new Vector();
        progressEventTargets.addElement( l );
    }

    /**
     * Remove a ProgressEvent listener.
     * <P>
     * @param l listener to remove
     */
    public synchronized void removeProgressListener( ProgressListener l ) {
        if ( progressEventTargets != null && progressEventTargets.size() > 0 )
            progressEventTargets.removeElement(l);
    }

    /**
     * Get all listeners for ProgressEvent.
     * <P>
     * @return array of listeners
     */
    public ProgressListener[] getProgressListeners() {
        if ( progressEventTargets != null )
            return ( ProgressListener[] ) progressEventTargets.toArray();
        else
            return null;
    }
    
    /**
     * Notify the ProgressEvent listeners about the training progress.
     * <P>
     * @param percent percentage of training progress
     */
    protected void notifyProgressEventTargets( int percent ) {
        Vector list;
        if ( progressEventTargets != null ) {            
            if( percent < 0 ) 
                percent = 0;
            else if ( percent > 100 )
                percent = 100;
            ProgressEvent event = new ProgressEvent( this, percent );
            synchronized( this ) {
                list = ( Vector ) progressEventTargets.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                ProgressListener listener = ( ProgressListener) list.elementAt( i );
                listener.progressMade( event );
            }
        }
    }
    
    /**
     * Add a listener for TrainingEvent.
     * <P>
     * @param l listener to add
     */
    public synchronized void addTrainingListener( TrainingListener l ) {
        if ( trainingEventTargets == null )
            trainingEventTargets = new Vector();
        trainingEventTargets.addElement( l );
    }

    /**
     * Remove a TrainingEvent listener.
     * <P>
     * @param l listener to remove
     */
    public synchronized void removeTrainingListener( TrainingListener l ) {
        if ( trainingEventTargets != null && trainingEventTargets.size() > 0 )
            trainingEventTargets.removeElement(l);
    }
    
    /**
     * Get all listeners for TrainingEvent.
     * <P>
     * @return array of listeners
     */
    public TrainingListener[] getTrainingListeners() {
        if ( trainingEventTargets != null )
            return ( TrainingListener[] ) trainingEventTargets.toArray();
        else
            return null;
    }
    
    /**
     * Notify the TrainingEvent listeners about the training status. Possible values are:
     * <P>
     * TrainingEvent.TRAINING_INITIALIZED: training is initialized
     * TrainingEvent.TRAINING_STARTED:     training has started
     * TrainingEvent.TRAINING_STOPED:      training has stoped
     * <P>
     * @param msgID message to send
     */
    protected void notifyTrainingEventTargets( int msgID ) {
        Vector list;
        if ( trainingEventTargets != null ) {
            TrainingEvent event = new TrainingEvent( this, msgID );
            synchronized( this ) {
                list = ( Vector ) trainingEventTargets.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                TrainingListener listener = ( TrainingListener) list.elementAt( i );
                listener.trainingActionPerformed( event );
            }
        }
    }
    
    /**
     * Add a listener for PropertyChangeEvent.
     * <P>
     * @param l listener to add
     */
    public void addPropertyChangeListener( PropertyChangeListener l ) {
        changes.addPropertyChangeListener( l );
    }
    
    /**
     * Remove a PropertyChangeEvent listener.
     * <P>
     * @param l listener to remove
     */
    public void removePropertyChangeListener ( PropertyChangeListener l ) {
        changes.removePropertyChangeListener( l );
    }

    /**
     * Get a string representation of the bean.
     * <P>
     * @return representation string
     */
    public String toString() {
        StringBuffer retVal = new StringBuffer( "Neural Network Bean\n\n" );
        retVal.append( "Description:\t" ).append( description );
        retVal.append( "\nTraining Function:\t" ).append( trainingFunction );
        retVal.append( "\nMax Training Cycles:\t").append( maxTrainingCycles );
        retVal.append( "\nFitness Function:\t" ).append( fitnessFunction );
        retVal.append( "\n" ).append( nn );
        return retVal.toString();
    }
    
    // get the name of a activation function
    private String getActFunctionName( ActFunction value ) {
        String retVal = null;
        String tmp = null;
        Enumeration keys = activationFunctions.keys();
        while ( keys.hasMoreElements() ) {                
            tmp = ( String ) keys.nextElement();
            if ( ( ( Class ) activationFunctions.get( tmp ) ).isInstance( value ) )
                retVal = tmp;
        }
        return retVal;
    }
}
