package edu.cosy.netgen.ann;

import java.io.*;
import java.util.*;

public class BP extends Training {
        
    protected ConnectionMatrix conMatrix;
    
    // amount that the weight matrix should be changed
    private double changeWeightMatrix[][];

    // if momentum is used t-1 matrix gets stored, offline
    private double changeWeightMatrixTempOffline[][];

    // if momentum is used t-1 matrixes get stored, online
    private double changeWeightMatrixTempOnlineContainer[][][];

    // trained cycles: either max. cycle is reached or error is less than the given
    private int trainedCycles;
    
    // contains the output of all neurons
    private double outputvek[];

    // contains the input of all neurons
    private double netvek[];

    // contains the delta of hidden+output neurons, 
    // same length as netvek and outputvek, input spaces contains garbage
    private double deltavek[];
        
    public BP() {
        super();
        params = new TrainingParam[ 3 ];
        
        params[ 0 ] = new TrainingParam( Double.class );
        params[ 0 ].setDescription( "learningRate" );
        params[ 0 ].setValues( new Double( 0.2 ), new Double( 0.0 ), new Double( 1.0 ) );

        params[ 1 ] = new TrainingParam( Double.class );
        params[ 1 ].setDescription( "momentum" );
        params[ 1 ].setValues( new Double( 0.8 ), new Double( 1.0 ), new Double( 2.0 ) );

        params[ 2 ] = new TrainingParam( Boolean.class );
        params[ 2 ].setDescription( "offline" );
        params[ 2 ].setValue( new Boolean( true ) );
    }
    
    public int getValidNetType() {
        return NeuralNet.FEEDFORWARD;
    }       

    public boolean init( NeuralNet nn, PatternSet trainSet ) {
        boolean retVal = false;     
        if ( nn instanceof FeedForwardNet ) {
            // set net
            this.nn = nn;
            conMatrix = nn.getConMatrix();
            
            // set training sets
            this.trainingSet = trainSet;
    
            matrixWidth = nn.getIn() + nn.getHidden() + nn.getOut() + 1;
    
            //contains the values of the change request
            changeWeightMatrix = new double[ matrixWidth ][ matrixWidth ];
    
            //contains the changeMatrix(t-1) for offline learning
            changeWeightMatrixTempOffline = new double[ matrixWidth ][ matrixWidth ];
    
            outputvek = new double[ matrixWidth ];
            netvek = new double[ matrixWidth ];
            deltavek = new double[ matrixWidth ];
    
            //contains the changeMatrixes(t-1) for online learning
            changeWeightMatrixTempOnlineContainer = new double[ trainingSet.getSize() ][ matrixWidth ][ matrixWidth ];
            retVal = true;
        }
        return retVal;
    }
        
    //the deltavek gets initialized
    private void initBackwardCycle( int patPos ) {
        
        //the smallest number gets assigned to distinguish between valid values and init values
        for ( int i = 0; i < matrixWidth; i++ )
            deltavek[ i ] = Double.longBitsToDouble( 0x1L );
            
        //deltavek init for output neurons, these values exist always
        for ( int i = 1 + nn.getIn() + nn.getHidden(); i < nn.getIn() + nn.getHidden() + nn.getOut() + 1; i++ ) {
            deltavek[ i ] = nn.getActFunctions()[ i ].actFunctionDerivation( outputvek[ i ] ) * ( trainingSet.getOutPatternArray()[ patPos ][ i - ( 1 + nn.getIn() + nn.getHidden() ) ] - outputvek[ i ] );
        }
    }

    //the error of the whole net is computed
    private void backwardCycle() {
        for ( int k = 1 + nn.getIn(); k < 1 + nn.getIn() + nn.getHidden(); k++ ) {
            backward( k );
        }
    }

    //a backward step for one neuron, if a value of a previous neuron is missing
    //the method gets executed recursively
    private void backward( int neuron ) {
        
        //determines that work should be done
        if ( deltavek[ neuron ] == Double.longBitsToDouble( 0x1L ) ) {
            
            //determines the pos. in a list which contains all outgoing neurons of a special neuron
            int outListPos = 0;
            
            //outgoing neurons for every neuron
            double outList[][];
            
            //row 0: neuron we are interested in, row 1: weights of con. between
            //interesting neuron and outgoing neurons
            outList = new double[ 2 ][ matrixWidth ];

            //for every column in the con. matrix
            for ( int k = 0; k < matrixWidth; k++ ) {
                
                //if there is a connection to an outgoing neuron
                if ( conMatrix.isConnection( neuron, k ) ) {
                    
                    //outListPos: position of outgoing neuron, k: outgoing neuron
                    outList[ 0 ][ outListPos ] = k;
                    outList[ 1 ][ outListPos ] = nn.getWeightMatrix().getWeight( neuron, k );
                    
                    //ready for the next outgoing neuron
                    outListPos++;
                    
                    //an other neuron is needed to compute the actual neuron
                    if ( outputvek[ k ] ==  Double.longBitsToDouble( 0x1L ) ) {
                        backward( k );
                    }
                }
            }

            //all weights of the outgoing neurons are multiplied with the corresponding
            //deltavek value, and the products are summed up
            double sum = 0;
            for ( int l = 0; l < outListPos; l++ ) {
                sum += deltavek[ ( int ) outList[ 0 ][ l ] ] * outList[ 1 ][ l ];
            }
            
            //a new value for the deltavek is now computed
            deltavek[ neuron ] = nn.getActFunctions()[ neuron ].actFunctionDerivation( outputvek[ neuron ] ) * sum;

            //if there are no outgoing neurons or the outgoing neuron was killed by a marker
            if ( ( outListPos == 0 ) || ( conMatrix.getMarkers() != null && conMatrix.getMarkers()[ neuron ] ) )
                deltavek[ neuron ] = 0;
        }
    }

    //a change request matrix is build (offline)
    private void createChangeMatrix() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) ) {
                    if ( getBooleanValue( params[ 2 ].getValue() ) )
                        changeWeightMatrix[ i ][ j ] += getDoubleValue( params[ 0 ].getValue() ) * outputvek[ i ] * deltavek[ j ];
                    else
                        changeWeightMatrix[ i ][ j ] = getDoubleValue( params[ 0 ].getValue() ) * outputvek[ i ] * deltavek[ j ];   
                }
            }
        }
    }

    //the change matrix is saved in a temp matrix, it represent the t-1 matrix
    //in the next step, offline
    private void makeChangeMatrixTempOffline() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) ) {
                    changeWeightMatrixTempOffline[ i ][ j ] = changeWeightMatrix[ i ][ j ];
                }
            }
        }
    }

    //the actual change matrix is reset
    private void resetChangeMatrixOffline() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                changeWeightMatrix[i][j] = 0;
            }
        }
    }

    //the weight matrix gets changed, momentum matrixes are only used when momentum should be used, online
    private void changeWeightMatrixOnline( int patPo ) {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) ) {
                    if ( getDoubleValue( params[ 1 ].getValue() ) != 0.0 )
                        nn.getWeightMatrix().setWeight( i, j, nn.getWeightMatrix().getWeight( i, j ) + changeWeightMatrix[ i ][ j ] + getDoubleValue( params[ 1 ].getValue() ) * changeWeightMatrixTempOnlineContainer[ patPo ][ i ][ j ] );
                    else
                        nn.getWeightMatrix().setWeight( i, j, nn.getWeightMatrix().getWeight( i, j ) + changeWeightMatrix[ i ][ j ] );
                }
            }
        }
    }
    
    //the weight matrix gets changed, momentum matrixes are only used when momentum should be used, offline
    private void changeWeightMatrixOffline() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) ) {
                    if( getDoubleValue( params[ 1 ].getValue() ) != 0 )
                        nn.getWeightMatrix().setWeight( i, j, nn.getWeightMatrix().getWeight( i, j ) + changeWeightMatrix[ i ][ j ] + getDoubleValue( params[ 1 ].getValue() ) * changeWeightMatrixTempOffline[ i ][ j ] );
                    else
                        nn.getWeightMatrix().setWeight( i, j, nn.getWeightMatrix().getWeight( i, j ) + changeWeightMatrix[ i ][ j ] );
                }
            }
        }
    }

    /**
     * Trains the net, offline: weight update after all patterns, online: weight update after each pattern
     * @return how many cycles the net was trained
     */
    public int train( ANN bean ) {
        double stopValue = 1.0;
        FP fp = new FP( ( FeedForwardNet ) nn, trainingSet );
        trainedCycles = 0;        
        double percent = 0.0;        
        Vector helpin;

        while ( true ) {
            if ( ( bean.getMaxTrainingCycles() > 0 && bean.getMaxTrainingCycles() == trainedCycles ) ||
                 ( bean.getStopError() > 0.0 && bean.getStopError() > stopValue ) )
                break;
            if ( bean != null && bean.getMaxTrainingCycles() > 0 ) {
                percent = ( ( double ) trainedCycles ) / ( ( double ) bean.getMaxTrainingCycles() ) * 100;
                bean.notifyProgressEventTargets( ( int ) percent );
            }

            //for all patterns
            for ( int patPo = 0; patPo < trainingSet.getSize(); patPo++ ) {
                if ( !trainingSet.isMarker( patPo ) ) {
                
                    //a forward step takes place
                    helpin = fp.forw( patPo );
    
                    //result of the forward step
                    outputvek = ( double[] ) helpin.elementAt( 0 );
                    
                    //result of the forward step
                    netvek = ( double[] ) helpin.elementAt( 1 );
                    
                    //make backward step for the actual pattern
                    initBackwardCycle( patPo );
                    backwardCycle();
    
                    //how many should the weight matrix be changed?
                    createChangeMatrix();
                    
                    if ( !getBooleanValue( params[ 2 ].getValue() ) ) {
                        //if momentum should be used temp matrixes for each pattern are stored
                        if ( getDoubleValue( params[ 1 ].getValue() ) != 0.0 )
                              System.arraycopy( changeWeightMatrix, 0, changeWeightMatrixTempOnlineContainer[ patPo ], 0, changeWeightMatrix.length );
                        changeWeightMatrixOnline( patPo );
                    }
                }
           }
           if ( getBooleanValue( params[ 2 ].getValue() ) ) {
                
                //change the weight matrix
                changeWeightMatrixOffline();
            
                //momentum should be used if true
                if ( getDoubleValue( params[ 1 ].getValue() ) != 0.0 )
                    makeChangeMatrixTempOffline();
                
                //set the change matrix to zero
                resetChangeMatrixOffline();
            }
            
            //stop training when error is small enough
            stopValue = bean.getFitness();
            trainedCycles++;
        }
        return trainedCycles;
    }
    
    private final double getDoubleValue( Object value ) {
        return ( ( Double ) value ).doubleValue();
    }

    private final boolean getBooleanValue( Object value ) {
        return ( ( Boolean ) value ).booleanValue();
    }
}
