package edu.cosy.netgen.ann;

/**
 * Provides a connection matrix for neural networks.
 */
public class ConnectionMatrix implements java.io.Serializable {
    
    /**
     * Net type for a none connected network.
     */
    public final static int CREATE_NONE = 200;

    /**
     * Net type for a normal three layer network.
     */
    public final static int CREATE_THREE_LAYER = 201;
    
    /**
     * Net type for a full connected network.
     */
    public final static int CREATE_FULL = 202;
    
    /**
     * Net type for a random connected network.
     */
    public final static int CREATE_RANDOM = 203;
    
    /**
     * Net type for a net with bias connections only.
     */
    public final static int CREATE_BIAS_ONLY = 204;
    
    /**
     * Connections.
     */
    protected boolean[][] connections;
    
    /**
     * Markers for neurons. If it is true, the connection does not exists.
     * If markers is null, the nothing happens.
     */
    protected boolean[] markers = null;
    
    /**
     * Number of input, hidden and output neurons.
     */
    protected int input, hidden, output;
    
    /**
     * Connection style.
     */
    protected int connectionStyle = CREATE_THREE_LAYER;
    
    /**
     * Create a new connection matrix with one bias neuron, none of input, hidden or output.
     */
    public ConnectionMatrix() {
        this( 0, 0, 0, CREATE_THREE_LAYER );
    }
    
    /**
     * Copy constructor for the class.
     * <P>
     * @param cm connection matrix to copy
     */
    public ConnectionMatrix( ConnectionMatrix cm ) {
        super();
        connections = new boolean[ cm.connections.length ][ cm.connections.length ];
        System.arraycopy( cm.connections, 0, connections, 0, connections.length );
        if ( cm.markers != null ) {
            markers = new boolean[ cm.markers.length ];
            System.arraycopy( cm.markers, 0, markers, 0, markers.length );
        }
        this.input = cm.input;
        this.hidden = cm.hidden;
        this.output = cm.output;
        this.connectionStyle = cm.connectionStyle;
    }
    
    /**
     * Create a new connection matrix. The size is +1 because of possible connections to bias neurons.
     * <P>
     * @param in number of input neurons
     * @param hidden number of hidden neurons
     * @param out number of output neurons
     * @param style connection style of the matrix
     */
    public ConnectionMatrix( int in, int hidden, int out, int style ) {
        super();        
        this.input = in;
        this.hidden = hidden;
        this.output = out;
        connections = new boolean[ 1 + input + hidden + output ][ 1 + input + hidden + output ];
        markers = null;
        setConnectionStyle( style );
    }
        
    /**
     * Set the connection style for the matrix.
     * <P>
     * @param style connetction style
     */
    public void setConnectionStyle( int style ) throws java.lang.IllegalArgumentException {
        if ( connections != null ) {
             switch ( style ) {                
                case CREATE_THREE_LAYER : {
                    connectForwardThreeLayer();
                    break;
                }
                case CREATE_FULL : {
                    connectForwardFull();
                    break;
                }
                case CREATE_RANDOM : {
                    connectForwardRandom();
                    break;                    
                }
                case CREATE_BIAS_ONLY : {
                    for ( int j = 1 + input; j < size() + 1; j++ )
                        connections[ 0 ][ j ] = true;
                    break;                    
                }
                case CREATE_NONE : break;
                default: {
                    throw new java.lang.IllegalArgumentException();
                }
            }
        }
    }
    
    /**
     * Get the current connection style of the matrix.
     * <P>
     * @return connection style
     */
    public int getConnectionStyle() {
        return connectionStyle;
    }

    /**
     * Get the size of the matrix -1 for bias connections.
     * <P>
     * @return matrix size
     */
    public int size() {
        int retVal = 0;
        if ( connections != null )
            retVal = connections.length - 1;
        return retVal;
    }
    
    /**
     * Set a connection from a source to a destination neuron.
     * <P>
     * @param s source neuron
     * @param d destination neuron
     */
    public void setConnection( int s, int d ) {
        if ( connections != null && connections.length > s && connections.length > d )
            connections[ s ][ d ] = true;        
    }
    
    /**
     * Checking of a connection.
     * <P>
     * @param s source neuron
     * @param d destination neuron
     * @return true, if a connection exists
     */
    public boolean isConnection( int s, int d ) {
        if ( connections != null && connections.length > s && connections.length > d )
            if ( markers != null && ( markers[ s ] || markers[ d ] ) )
                return false;
            else
                return connections[ s ][ d ];
        else
            return false;
    }
    
    /**
     * Checking if there is any connection to this neuron.
     * <P>
     * @param value neuron to check
     * @return true, if there is any connection
     */
    public boolean isAnyConnection( int value ) {
        if ( connections != null && connections.length > value ) {
            for ( int i = 1; i < connections.length; i++ ) {
                if ( isConnection( i, value ) )
                    return true;                
            }
        }
        return false;
    }

    /**
     * Count the connections without the bias ones.
     * <P>
     * @return number of connections
     */
    public int countConnections() {
        int retVal = 0;
        for ( int i = 1; i < connections.length; i++ ) {
            for ( int j = 1; j < connections.length; j++ ) {
                if ( connections[ i ][ j ] )
                    retVal++;
            }
        }
        return retVal;
    }
     
    /**
     * Set the neuron markers. If it is null or the length is unequal to the number of neurons,
     * a java.lang.IllegalArgumentException is thrown.
     * <P>
     * @param values array of markers
     */
    public void setMarkers( boolean[] values ) throws java.lang.IllegalArgumentException {
        if ( connections != null && connections.length == values.length )
            markers = values;
        else
            throw new java.lang.IllegalArgumentException();
    }
       
    /**
     * Reset the markers to null.
     */ 
    public void resetMarkers() {
        markers = null;
    }
    
    /**
     * Get the markers of the connection matrix. If the return value is null, than no markers are set.
     * <P>
     * @return array of markers
     */
    public boolean[] getMarkers() {
        return markers;
    }
    
    /**
     * Checking of the matrix for connections to input neurons or recurrent connections.
     * False connections will be deleted.
     * <P>
     */
    public void checkForward() {
        if ( connections != null ) {
            int size = 1 + input + hidden + output;
        
            //sets bias in the con. matrix
            for ( int j = 1 + input; j < size; j++ )
                connections[ 0 ][ j ] = true;

            //deletes recurrent connections
            for ( int i = 1; i < size; i++ ) {
                for ( int j = 1; j < size; j++ ) {
                    if ( i >= j )
                        connections[ i ][ j ] = false;
                }
            }

            //deletes connections to input units
            for ( int i = 1; i < size; i++ ) {
                for ( int j = 1; j < 1 + input; j++ ) {
                    connections[ i ][ j ] = false;
                }
            }
        }
    }

    /**
     * Get a String representation of the matrix.
     * <P>
     * @return String representation
     */
    public String toString() {
        StringBuffer retVal = new StringBuffer( "Connection Matrix:\n\nConnection Style:" );
        if ( connections != null ) {
             switch ( connectionStyle ) {                
                case CREATE_THREE_LAYER: {
                    retVal.append( "Three Layer FeedForward\n" );
                    break;
                }
                case CREATE_FULL: {
                    retVal.append( "Full Connected FeedForward\n" );
                    break;
                }
                case CREATE_RANDOM: {
                    retVal.append( "Random Connected FeedForward\n" );
                    break;                    
                }
                default:
                    retVal.append( "unknown\n" );
            }            
            for ( int i = 0; i < connections.length; i++ ) {
                if ( i == 0 )
                    retVal.append( "B(" + i + ")\t" );
                else if ( i < input + 1 )
                    retVal.append( "I(" + i + ")\t" );
                else if ( i < input + hidden + 1 )
                    retVal.append( "H(" + i + ")\t" );
                else
                    retVal.append( "O(" + i + ")\t" );
                for ( int j = 0; j < connections.length; j++ ) {
                    if ( connections[ i ][ j ] )
                        retVal.append( "1  " );
                    else
                        retVal.append( "0  " );
                }
                retVal.append( "\n" );
            }
        }
        return retVal.toString();
    }
       
    // Create a full connected FeedForward Net 
    private void connectForwardFull() {
        if ( connections != null ) {
            int size = 1 + input + hidden + output;

            //init bias and connections from input to hidden and output
            for ( int i = 0; i < 1 + input; i++ ) {
                for ( int j = 1 + input; j < size; j++ ) {
                    connections[ i ][ j ] = true;
                }
            }
            
            //init connections between hidden
            for ( int i = 1 + input; i < 1 + input + hidden; i++ ) {
                for ( int j = i + 1; j < size; j++ ) {
                    connections[ i ][ j ] = true;
                }
            }
        }
    }

    // Create a random connected FeedForward Net
    private void connectForwardRandom() {
        if ( connections != null ) {
            int size = 1 + input + hidden + output;
    
            //init bias
            for ( int j = 1 + input; j < size; j++ )
                connections[ 0 ][ j ] = true;

            //init connections from input to hidden and output
            for ( int i = 1; i < 1 + input; i++ ) {
                for ( int j = 1 + input; j < size; j++ ) {
                    if ( Math.random() < 0.5 )
                        connections[ i ][ j ] = true;
                }
            }

            //init connections between hidden
            for ( int i = 1 + input; i < 1 + input + hidden; i++ ) {
                for ( int j = i + 1; j < size; j++ ) {
                    if ( Math.random() < 0.5 )
                        connections[ i ][ j ] = true;
                }
            }
        }
    }

    // Create a three layer FeedForward Net
    private void connectForwardThreeLayer() {
        if ( connections != null ) {
            int size = 1 + input + hidden + output;
            
            //init bias
            for ( int j = 1 + input; j < size; j++ ) {
                connections[ 0 ][ j ] = true;
            }
            
            //init connection from input to hidden
            for ( int i = 0; i < 1 + input; i++ ) {
                for ( int j = 1 + input; j < 1 + input + hidden; j++ ) {
                    connections[ i ][ j ] = true;
                }
            }
            
            //init connections from hidden to output
            for ( int i = 1 + input; i < 1 + input + hidden; i++ ) {
                for ( int j = 1 + input + hidden; j < size; j++ ) {
                    connections[ i ][ j ] = true;
                }
            }
        }
    }
}
