package edu.cosy.netgen.ann;

import java.io.*;
import java.util.*;

/**
 * Provides FeedForward calculations.
 */
public class FP implements java.io.Serializable {

    //net structure
    private FeedForwardNet ffn;

    //act. func. array
    private ActFunction actFunction[];

    protected ConnectionMatrix conMatrix;

    //weight matrix
    private WeightMatrix weightMatrix;

    /**
     * Contains the output of all neurons
     */
    protected double outputvek[];

    //contains the input of all neurons
    private double netvek[];

    //net result for a input patternset
    private double resultMatrix[][];

    // matrix width
    private int matrixWidth = 0;
    
    // calculation pattern set
    private PatternSet calculationSet;
    
    /**
     * Creates a new forward propagation instance
     * @param ffn feed forward net structure
     * @param calcSet calculation pattern set
     */
    public FP( FeedForwardNet ffn, PatternSet calcSet ) {
        this.ffn = ffn;
        conMatrix = ffn.getConMatrix();
        this.calculationSet = calcSet;

        this.matrixWidth = ffn.getIn() + ffn.getOut() + ffn.getHidden() + 1;

        weightMatrix = ffn.getWeightMatrix();

        //create the desired act. func. for each neuron
        actFunction = ffn.getActFunctions();

        outputvek = new double[ matrixWidth ];
        netvek = new double[ matrixWidth ];

        resultMatrix = new double[ calcSet.getSize() ][ ffn.getOut() ];
    }

    /**
     * Get the result for an input patternset
     * @return result matrix (row: result for a pattern, column: output neurons)
     */
    public double[][] getResultMatrix() {
        return resultMatrix;
    }

    /**
     * Makes a forward step in the net for one neuron, if for the computation another neuron is necessary the method gets executed again
     * @param neuron forward for which neuron
     */
    public void forward( int neuron ) {

        //if no value was computed for neuron
        if ( outputvek[ neuron ] == Double.longBitsToDouble( 0x1L ) ) {
            //determines the pos. in a list which contains all ingoing neurons of a special neuron
            int inListPos = 0;
            
            //ingoing neurons for every neuron
            double inList[][];
            
            //row 0: neuron we are interested in, row 1: weights of con. between
            //interesting neuron and ingoing neurons
            inList = new double[ 2 ][ matrixWidth ];
            
            //for all ingoing neurons
            for ( int k = 0; k < matrixWidth; k++ ) {
                //if there is a connection
                if ( conMatrix.isConnection( k, neuron ) ) {
                    
                    //inListPos: position of ingoing neuron, k: ingoing neuron
                    inList[ 0 ][ inListPos ] = k;
                    inList[ 1 ][ inListPos ] = weightMatrix.getWeight( k, neuron );
                    
                    //ready for the next ingoing neuron
                    inListPos++;
                    
                    //an other neuron is needed to compute the actual neuron
                    if ( outputvek[ k ] == Double.longBitsToDouble( 0x1L ) ) {
                        int neuronRec = k;
                        forward( neuronRec );
                    }
                }
            }

            //all weights of the ingoing neurons are multiplied with the corresponding
            //outputvek value, and the products are summed up
            double sum = 0;
            for ( int l = 0; l < inListPos; l++ ) {
                sum += outputvek[ ( int ) inList[ 0 ][ l ] ] * inList[ 1 ][ l ];
            }
            
            //a new value for the outputvek is now computed
            outputvek[ neuron ] = actFunction[ neuron ].actFunction( sum );
            
            //a new value for the netvek is now computed
            netvek[ neuron ] = sum;

            //if there are no ingoing neurons or the ingoing neuron was killed by a marker
            if ( ( inListPos == 0 ) || ( conMatrix.getMarkers() != null && conMatrix.getMarkers()[ neuron ] ) )
                outputvek[ neuron ] = 0;
        }
    }

    /**
     * Initialize forward cycle
     * @param patPos position of the pattern
     */
    protected void initForwardCycle( int patPos ) {
        
        //init values to distinguish between computed neurons and not computed neurons
        for ( int i = 0; i < matrixWidth; i++ )
            outputvek[ i ] = Double.longBitsToDouble( 0x1L );
        for ( int i = 3; i < matrixWidth; i++ )
            netvek[ i ] = Double.longBitsToDouble( 0x1L );
        //bias
        outputvek[ 0 ] = 1;
        
        //bias
        netvek[ 0 ] = 1;
        
        //value of input neurons are always present
        for ( int k = 1; k < 1 + ffn.getIn(); k++ ) {
            outputvek[ k ] = actFunction[ k ].actFunction( calculationSet.getInPatternArray()[ patPos ][ k - 1 ] );
            netvek[ k ] = calculationSet.getInPatternArray()[ patPos ][ k - 1 ];
            
            //input neuron is killed by a marker
            if ( conMatrix.getMarkers() != null && conMatrix.getMarkers()[ k ] ) {
                outputvek[ k ] = 0;
                netvek[ k ] = 0;
            }
        }
    }

    /**
     * The input pattern is propagated through the whole net
     */
    protected void forwardCycle() {
        for ( int k = 1 + ffn.getIn(); k < matrixWidth; k++ ) {
            forward( k );
        }
    }

    /**
     * Makes a forward step for the whole net and for all patterns,
     * a net gets calculated
     */
    protected void forwardStep() {
        
        //for all patterns
        for ( int i = 0; i < calculationSet.getSize(); i++ ) {
            initForwardCycle( i );
            forwardCycle();
            
            //fill the result matrix
            for ( int j = 1 + ffn.getIn() + ffn.getHidden(); j < matrixWidth; j++ )
                resultMatrix[ i ][ j - ( 1 + ffn.getIn() + ffn.getHidden() ) ] = outputvek[ j ];
        }
    }

    /**
     * Necessary for bp, rprop to compute a forward cycle
     * @param help help vector
     * @param patPos position of the pattern
     * @return new vector
     */
    protected Vector forw( int patPos ) {
        Vector vec = new Vector();
        initForwardCycle( patPos );
        forwardCycle();
        vec.addElement( outputvek );
        vec.addElement( netvek );
        return vec;
    }

    //square for double values
    private double sqr(double d) {
        return d * d;
    }
}
