package edu.cosy.netgen.ann;

import java.io.*;
import java.util.*;

/**
 * Provides a RProp Training Algorithm for FeedForward Networks.
 */
public class Rprop extends Training {

    protected ConnectionMatrix conMatrix;
    
    //weight matrix t-1
    private WeightMatrix weightMatrixTemp;

    //change weight matrix
    private double changeWeightMatrix[][];

    //change weight matrix t-1
    private double changeWeightMatrixTemp[][];

    //s matrix
    private double sMatrix[][];

    //s matrix t-1
    private double sMatrixTemp[][];

    //delta matrix
    private double deltaMatrix[][];

    //delta matrix t-1
    private double deltaMatrixTemp[][];

    //contains the output of all neurons
    private double outputvek[];

    //contains the input of all neurons
    private double netvek[];

    //contains the delta of hidden+output neurons, same length as netvek
    //and outputvek, input spaces contains garbage
    private double deltavek[];

    public Rprop() {
        super();
        params = new TrainingParam[ 3 ];
    
        params[ 0 ] = new TrainingParam( Double.class );
        params[ 0 ].setDescription( "nMinus" );
        params[ 0 ].setValues( new Double( 0.5 ), new Double( 0.0 ), new Double( 1.0 ) );

        params[ 1 ] = new TrainingParam( Double.class );
        params[ 1 ].setDescription( "nPlus" );
        params[ 1 ].setValues( new Double( 1.2 ), new Double( 1.0 ), new Double( 2.0 ) );

        params[ 2 ] = new TrainingParam( Double.class );
        params[ 2 ].setDescription( "delta" );
        params[ 2 ].setValues( new Double( 1.0 ), new Double( 0.0 ), new Double( 1.0 ) );
    }
    
    public int getValidNetType() {
        return NeuralNet.FEEDFORWARD;
    }       

    public boolean init( NeuralNet nn, PatternSet trainSet ) {
        boolean retVal = false;     
        if ( nn instanceof FeedForwardNet ) {
            this.nn = nn;
            this.trainingSet = trainSet;
            
            conMatrix = nn.getConMatrix();
            matrixWidth = nn.getIn() + nn.getHidden() + nn.getOut() + 1;            
            weightMatrixTemp = new WeightMatrix( nn.getWeightMatrix() );
    
            sMatrix = new double[ matrixWidth ][ matrixWidth ];
            sMatrixTemp = new double[ matrixWidth ][ matrixWidth ];
            changeWeightMatrix = new double[ matrixWidth ][ matrixWidth ];
            changeWeightMatrixTemp = new double[ matrixWidth ][ matrixWidth ];
            deltaMatrix = new double[ matrixWidth ][ matrixWidth ];
            deltaMatrixTemp = new double[ matrixWidth ][ matrixWidth ];
    
            //init all temps
            for ( int i = 0; i < matrixWidth; i++ ) {
                for (int j = 0; j < matrixWidth; j++) {
                    deltaMatrixTemp[i][j] = 0.1;
                    sMatrixTemp[i][j] = 0;
                    changeWeightMatrixTemp[i][j] = 0;
                }
            }
    
            outputvek = new double[ matrixWidth ];
            netvek = new double[ matrixWidth ];
            deltavek = new double[ matrixWidth ];
            
            retVal = true;
        }
        return retVal;
    }
    
    //the deltavek gets initialized
    private void initBackwardCycle( int patPos ) {

        //init values differ from real values
        for ( int i = 0; i < matrixWidth; i++ )
            deltavek[ i ] = Double.longBitsToDouble( 0x1L );

        //fill deltavek with values from output neurons
        for ( int i = 1 + nn.getIn() + nn.getHidden(); i < matrixWidth; i++ ) {
            deltavek[ i ] = nn.getActFunctions()[ i ].actFunctionDerivation( outputvek[ i ] )
                            * ( trainingSet.getOutPatternArray()[ patPos ][ i - ( 1 + nn.getIn() + nn.getHidden() ) ] - outputvek[ i ] );
        }
    }

    //the error of the whole net is computed
    private void backwardCycle() {
        for ( int k = 1 + nn.getIn(); k < 1 + nn.getIn() + nn.getHidden(); k++ )
            backward( k );
    }

    //backward cycle, the same as in BP.java
    private void backward( int neuron ) {
        if ( deltavek[ neuron ] == Double.longBitsToDouble( 0x1L ) ) {
            int outListPos = 0;

            double outList[][];
            outList = new double[2][ matrixWidth ];

            for ( int k = 0; k < matrixWidth; k++ ) {
                if ( conMatrix.isConnection( neuron, k ) ) {
                    outList[ 0 ][ outListPos ] = k;
                    outList[ 1 ][ outListPos ] = nn.getWeightMatrix().getWeight( neuron, k );
                    outListPos++;

                    if ( outputvek[ k ] == Double.longBitsToDouble( 0x1L ) )
                        backward(k);
                }
            }

            double sum = 0;
            for ( int l = 0; l < outListPos; l++ ) {
                sum += deltavek[ ( int ) outList[ 0 ][ l ] ] * outList[ 1 ][ l ];
            }

            deltavek[ neuron ] = nn.getActFunctions()[ neuron ].actFunctionDerivation( outputvek[ neuron ] ) * sum;

            if ( ( outListPos == 0 ) || ( conMatrix.getMarkers() != null && conMatrix.getMarkers()[ neuron ] ) )
                deltavek[ neuron ] = 0;
        }
    }

    //S matrix gets built
    private void createSMatrix() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) ) {
                    sMatrix[ i ][ j ] += outputvek[ i ] * ( -deltavek[ j ] );
                }
            }
        }
    }

    //actual values get old and tired
    private void makeTemps() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) ) {
                    weightMatrixTemp.setWeight( i, j, nn.getWeightMatrix().getWeight( i, j ) );
                    deltaMatrixTemp[ i ][ j ] = deltaMatrix[ i ][ j ];
                    sMatrixTemp[ i ][ j ] = sMatrix[ i ][ j ];
                    changeWeightMatrixTemp[ i ][ j ] = changeWeightMatrix[ i ][ j ];
                }
            }
        }
    }

    //reset the actual s matrix
    private void resetSMatrix() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                sMatrix[ i ][ j ] = 0;
            }
        }
    }

    //delta matrix creation
    private void createDeltaMatrix() {
        
        //all rows
        for ( int i = 0; i < matrixWidth; i++ ) {
            
            //all columns
            for ( int j = 0; j < matrixWidth; j++ ) {
                
                //if there is a connection
                if ( conMatrix.isConnection( i, j ) ) {
                    
                    //look for the update formula in the netGen manual
                    if ( ( sMatrixTemp[ i ][ j ] * sMatrix[ i ][ j ] ) > 0 ) {
                        deltaMatrix[ i ][ j ] = deltaMatrixTemp[ i ][ j ] * getDoubleValue( params[ 1 ].getValue() );
                        if ( ( deltaMatrixTemp[ i ][ j ] * getDoubleValue( params[ 1 ].getValue() ) ) > 50 )
                            deltaMatrix[ i ][ j ] = 50;
                        if ( ( deltaMatrixTemp[ i ][ j ] * getDoubleValue( params[ 1 ].getValue() ) ) < 0.000001 )
                            deltaMatrix[ i ][ j ] = 0.000001;
                    }
                    if ( ( sMatrixTemp[ i ][ j ] * sMatrix[ i ][ j ] ) < 0 ) {
                        deltaMatrix[ i ][ j ] = deltaMatrixTemp[ i ][ j ] * getDoubleValue( params[ 0 ].getValue() );
                        if ( ( deltaMatrixTemp[ i ][ j ] * getDoubleValue( params[ 1 ].getValue() ) ) > 50)
                            deltaMatrix[ i ][ j ] = getDoubleValue( params[ 1 ].getValue() );
                        if ( ( deltaMatrixTemp[ i ][ j ] * getDoubleValue( params[ 1 ].getValue() ) ) < 0.000001 )
                            deltaMatrix[ i ][ j ] = 0.000001;
                    }
                    if ( ( sMatrixTemp[ i ][ j ] * sMatrix[ i ][ j ]) == 0)
                        deltaMatrix[ i ][ j ] = deltaMatrixTemp[ i ][ j ];
                }
            }
        }
    }

    //change weights matrix creation
    private void createChangeWeightMatrix() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) ) {
                    
                    //look for the update formula in the netGen manual
                    if ( ( sMatrixTemp[ i ][ j ] * sMatrix[ i ][ j ]) > 0 && sMatrix[ i ][ j ] > 0 )
                        changeWeightMatrix[ i ][ j ] = -deltaMatrix[ i ][ j ];
                    else if ( ( sMatrixTemp[ i ][ j ] * sMatrix[ i ][ j ] ) > 0 && sMatrix[ i ][ j ] < 0 )
                        changeWeightMatrix[ i ][ j ] = deltaMatrix[ i ][ j ];
                    else if ( ( sMatrixTemp[ i ][ j ] * sMatrix[ i ][ j ] ) < 0 ) {
                        changeWeightMatrix[ i ][ j ] = -changeWeightMatrixTemp[ i ][ j ];
                        sMatrix[ i ][ j ] = 0;
                    } else {
                        if ( sMatrix[ i ][ j ] >= 0 )
                            changeWeightMatrix[ i ][ j ] = -1 * deltaMatrix[ i ][ j ];
                        if ( sMatrix[ i ][ j ] < 0 )
                            changeWeightMatrix[ i ][ j ] = deltaMatrix[ i ][ j ];
                    }
                }
            }
        }
    }

    //changed the weight matrix, depends on change weight matrix
    private void createWeightMatrix() {
        for ( int i = 0; i < matrixWidth; i++ ) {
            for ( int j = 0; j < matrixWidth; j++ ) {
                if ( conMatrix.isConnection( i, j ) )
                    nn.getWeightMatrix().setWeight( i, j, weightMatrixTemp.getWeight( i, j ) + changeWeightMatrix[ i ][ j ] );
            }
        }
    }

    //realizes the training
    public int train( ANN bean ) {
        double stopValue = 1.0;
        FP fp = new FP( ( FeedForwardNet ) nn, trainingSet );
        int trainedCycles = 0;
        double percent = 0.0;
        Vector helpin;

        while ( true ) {
            if ( ( bean.getMaxTrainingCycles() > 0 && bean.getMaxTrainingCycles() == trainedCycles ) ||
                 ( bean.getStopError() > 0.0 && bean.getStopError() > stopValue ) )
                break;
            if ( bean != null && bean.getMaxTrainingCycles() > 0 ) {
                percent = ( ( double ) trainedCycles ) / ( ( double ) bean.getMaxTrainingCycles() ) * 100;
                bean.notifyProgressEventTargets( ( int ) percent  );
            }
            for ( int patPo = 0; patPo < trainingSet.getSize(); patPo++ ) {
                if ( !trainingSet.isMarker( patPo ) ) {
                    helpin = fp.forw( patPo );
                    outputvek = ( double[] ) helpin.elementAt( 0 );
                    netvek = ( double[] ) helpin.elementAt( 1 );
                    initBackwardCycle( patPo );
                    backwardCycle();
                    createSMatrix();
                }
            }
            createDeltaMatrix();
            createChangeWeightMatrix();
            createWeightMatrix();

            makeTemps();
            resetSMatrix();
            stopValue = bean.getFitness();
            trainedCycles++;
        }
        return trainedCycles;
    }
    
    private final double getDoubleValue( Object value ) {
        return ( ( Double ) value ).doubleValue();
    }
}
