package edu.cosy.netgen.ga;

import banda.genalg.Genotype;
import banda.genalg.BitChromosome;

import edu.cosy.netgen.ann.*;

/**
 * Abstract class for ANN-based genotypes.
 */
public abstract class ANNGenotype extends Genotype {
    
    /**
     * Extension of double values for encoding.
     * For Example: 0.5 -> 0.500000
     */
    protected final static int DOUBLE_CODING_EXT = 5;
    
    /**
     * Temporal copy of the original bean.
     */
    protected ANN bean = null;
    
    /**
     * Flag for evolution of the connection matrix.
     */
    protected boolean isCO = false;

    /**
     * Flag for evolution of the training parameters.
     * Only double and boolean values will be evolved.
     */
    protected boolean isTP = false;

    /**
     * Flag for evolution of the maximum training cycles.
     */
    protected boolean isCY = false;
    
    /**
     * Flag for evolution of activation functions.
     */    
    protected boolean isAC = false;
    
    /**
     * Flag for evolution of neuron markers.
     */
    protected boolean isMA = false;

    /**
     * Flag for evolution of the weight matrix.
     */
    protected boolean isWT = false;
    
    
    /**
     * Minimum weight value for the evolved weight matrix.
     */
    protected double minW = 0.0;

    /**
     * Maximum weight value for the evolved weight matrix.
     */
    protected double maxW = 0.0;

    /**
     * Pool of valid activation functions for input neurons.
     */
    protected String[] actFunctionPoolIn;

    /**
     * Pool of valid activation functions for hidden neurons.
     */
    protected String[] actFunctionPoolHidden;

    /**
     * Pool of valid activation functions for output neurons.
     */
    protected String[] actFunctionPoolOut;
    
    /**
     * Handle for the training pattern set.
     */
    protected transient PatternSet train = null;
    
    /**
     * Handle for the validation pattern set.
     */
    protected transient PatternSet validate = null;
    
    /**
     * Create a new genotype with a specific number of chromosomes.
     * <P>
     * @param c number of chromosomes
     */
    public ANNGenotype( int c ) {
        super( c );
    }
    
    /**
     * Copy constructor for this genotype.
     * <P>
     * @param genotype genotype to copy
     */
    public ANNGenotype( ANNGenotype genotype ) {
        super( genotype );
        bean = genotype.bean;
        isCO = genotype.isCO;
        isTP = genotype.isTP;
        isCY = genotype.isCY;
        isAC = genotype.isAC;
        isMA = genotype.isMA;
        isWT = genotype.isWT;
        minW = genotype.minW;
        maxW = genotype.maxW;
        actFunctionPoolIn = genotype.actFunctionPoolIn;
        actFunctionPoolHidden = genotype.actFunctionPoolHidden;
        actFunctionPoolOut = genotype.actFunctionPoolOut;
        train = genotype.train;
        validate = genotype.validate;
    }
    
    /**
     * Set the pattern sets for training of the bean.
     * <P>
     * @param pst training pattern set
     * @param psv validation pattern set
     */
    public void setPatternSets( PatternSet pst, PatternSet psv ) {
        train = pst;
        validate = psv;
    }
        
    /**
     * Get the fitness of the genotype.
     * 1.0 is the maximum, 0.0 the minimum.
     * <P>
     * @return fitness value
     */
    public double getRawFitness() {  
        return getRawFitness( null );
    }
    
    /**
     * Get the fitness of the genotype.
     * 1.0 is the maximum, 0.0 the minimum.
     * <P>
     * @param listener lsitener for training progress
     * @return fitness value
     */
    public double getRawFitness( edu.cosy.netgen.ann.ProgressListener listener ) {  
        double retVal = 0.0;
        ANN tmp = ( ANN ) getPhenotype();
        if ( tmp != null && train != null && validate != null ) {
            tmp.setTrainingPatternSet( train );
            tmp.setValidationPatternSet( validate );
            if ( listener != null )
                tmp.addProgressListener( listener );
            if ( !isWT )
                tmp.trainNet();
            retVal = 1 - tmp.getFitness();
        }
        return retVal;
    }

    /**
     * Get the bit number for encoding the connection matrix without the bias connections.
     * <P>
     * @param bean master bean
     * @return bit length
     */
    protected int getConMatrixBitLength( ANN bean ) {
        int retVal = 0;
        if ( bean != null ) {
            int tmp = bean.getIn() + bean.getHidden() + bean.getOut();
            retVal = tmp * tmp;
        }
        return retVal;
    }
    
    /**
     * Decode a connection matrix from the chromosome.
     * <P>
     * @param chrom bit chromosome
     * @param index start index in the chromosome
     * @return new connection matrix.
     */
    protected ConnectionMatrix decodeConMatrix( BitChromosome chrom, int index ) {
        ConnectionMatrix retVal = new ConnectionMatrix( bean.getIn(), bean.getHidden(), bean.getOut(), ConnectionMatrix.CREATE_BIAS_ONLY );
        for ( int i = 1; i <= retVal.size(); i++ ) {
            for ( int j = 1; j <= retVal.size(); j++ ) {
                if ( chrom.get( index ) )
                    retVal.setConnection( i, j );
                index++;
            }
        }
        return retVal;
    }
        
    /**
     * Get the bit number for encoding the training parameters.
     * Only double and boolean values will be encoded.
     * <P>
     * @param bean master bean
     * @return bit length
     */
    protected int getTrainingParamBitLength( ANN bean ) {
        int retVal = 0;
        if ( bean != null ) {
            TrainingParam[] params = bean.getTrainingFunctionParams();
            double tmp = 0.0;
            for ( int i = 0; i < params.length; i++ ) {
                if ( params[ i ] != null && Double.class == params[ i ].getType() ) {
                    tmp = ( ( Double ) params[ i ].getMaxValue() ).doubleValue() - ( ( Double ) params[ i ].getMinValue() ).doubleValue();
                    if ( tmp > 0.0 )
                        retVal += getDoubleBitLength( tmp );                        
                } else if ( params[ i ] != null && Boolean.class == params[ i ].getType() )
                    retVal += 1;
            }        
        }
        return retVal;
    }
    
    /**
     * Decode training parameters from the chromosome.
     * <P>
     * @param chrom bit chromosome
     * @param index start index in the chromosome
     * @return array of training parameters
     */
    protected TrainingParam[] decodeTrainingParams( BitChromosome chrom, int index ) {
        TrainingParam[] retVal = bean.getTrainingFunctionParams();
        int ci = index;
        double tmp = 0.0;
        double value = 0.0;
        double maxValue = 0.0;
        for ( int i = 0; i < retVal.length; i++ ) {
            if ( retVal[ i ] != null && Double.class == retVal[ i ].getType() ) {
                value = ( ( Double ) retVal[ i ].getMinValue() ).doubleValue();
                maxValue = ( ( Double ) retVal[ i ].getMaxValue() ).doubleValue();
                tmp = maxValue - value;
                if ( tmp > 0.0 ) {
                    ci = index;
                    index += getDoubleBitLength( tmp );
                    value += decodeDouble( chrom.getBinaryInteger( ci, index - 1 ), tmp );
                    if ( value > maxValue )
                        value = maxValue;
                    retVal[ i ].setValue( new Double( value ) );
                }                        
            } else if ( retVal[ i ] != null && Boolean.class == retVal[ i ].getType() ) {
                if ( chrom.get( index ) )
                    retVal[ i ].setValue( new Boolean( true ) );
                else
                    retVal[ i ].setValue( new Boolean( false ) );
                index++;
            } 
        }       
        return retVal;
    }

    /**
     * Get the bit number for encoding the maximum training cycle.
     * <P>
     * @param bean master bean
     * @return bit length
     */
    protected int getMaxTrainingCyclesBitLength( ANN bean ) {        
        int retVal = 0;
        if ( bean != null )
            retVal = Integer.toBinaryString( bean.getMaxTrainingCycles() ).length();    
        return retVal;
    }
    
    /**
     * Decode maximum training cycles from the chromosome.
     * <P>
     * @param chrom bit chromosome
     * @param index start index in the chromosome
     * @return maximum training cycles
     */
    protected int decodeMaxTrainingCycles( BitChromosome chrom, int index ) {
        int retVal = 0;
        int tmp = Integer.toBinaryString( bean.getMaxTrainingCycles() ).length();
        if ( tmp > 0 )
            retVal = chrom.getBinaryInteger( index, index + tmp - 1 );
        index += tmp;
        return retVal; 
    }
    
    /**
     * Get the bit number for encoding the neuron activation functions.
     * <P>
     * @param bean master bean
     * @return bit length
     */
    protected int getActFunctionBitLength( ANN bean ) {
        int retVal = 0;
        if ( bean != null )
            retVal = ( bean.getIn() + bean.getHidden() + bean.getOut() ) * Integer.toBinaryString( bean.getActivationFunctionNames().length ).length();
        return retVal;
    }
    
    /**
     * Decode neuron activation functions from the chromosome.
     * <P>
     * @param chrom bit chromosome
     * @param index start index in the chromosome
     * @return array of activation functions
     */
    protected String[] decodeActFunctions( BitChromosome chrom, int index ) {
        String[] retVal = bean.getActivationFunctionsByNames();
        int in = bean.getIn();
        int hidden = bean.getHidden();
        int out = bean.getOut();
        int ci = index;
        int cl = Integer.toBinaryString( bean.getActivationFunctionNames().length ).length();
        int tmp;
        
        if ( actFunctionPoolIn != null ) {
            for ( int i = 1; i < 1 + in; i++ ) {
                index += cl;
                tmp = chrom.getBinaryInteger( ci, index - 1 );
                
                // if the number is a valid pool index
                if ( tmp > -1 && tmp < actFunctionPoolIn.length )
                    retVal[ i ] = actFunctionPoolIn[ tmp ];
                ci = index;
            }
        }

        if ( actFunctionPoolHidden != null ) {
            for ( int i = 1 + in; i < 1 + in + hidden; i++ ) {
                index += cl;
                tmp = chrom.getBinaryInteger( ci, index - 1 );
                
                // if the number is a valid pool index
                if ( tmp > -1 && tmp < actFunctionPoolHidden.length )
                    retVal[ i ] = actFunctionPoolHidden[ tmp ];
                ci = index;
            }
        }
        
        if ( actFunctionPoolOut != null ) {
            for ( int i = 1 + in + hidden; i < 1 + in + hidden + out; i++ ) {
                index += cl;
                tmp = chrom.getBinaryInteger( ci, index - 1 );
                
                // if the number is a valid pool index
                if ( tmp > -1 && tmp < actFunctionPoolOut.length )
                    retVal[ i ] = actFunctionPoolOut[ tmp ];
                ci = index;
            }
        }

        return retVal;
    }
    
    /**
     * Get the bit number for encoding the neuron markers of the input and hidden neurons.
     * <P>
     * @param bean master bean
     * @return bit length
     */
    protected int getMarkersBitLength( ANN bean ) {
        int retVal = 0;
        if ( bean != null )
            retVal = bean.getIn() + bean.getHidden();
        return retVal;
    }
    
    /**
     * Decode neuron markers from the chromosome.
     * <P>
     * @param chrom bit chromosome
     * @param index start index in the chromosome
     * @return markers array
     */
    protected boolean[] decodeMarkers( BitChromosome chrom, int index ) {
        boolean[] retVal = bean.getNeuronMarkers();
        if ( retVal == null ) {
            retVal = new boolean[ 1 + bean.getIn() + bean.getHidden() + bean.getOut() ];
            retVal[ 0 ] = false;
            for ( int i = 1 + bean.getIn() + bean.getHidden(); i < retVal.length; i++ ) 
                retVal[ i ] = false;
        }
        for ( int i = 1; i < 1 + bean.getIn() + bean.getHidden(); i++ ) {
            retVal[ i ] = chrom.get( index );
            index++;
        }
        return retVal;
    }
    
    /**
     * Get the bit number for encoding a new weight matrix.
     * <P>
     * @param bean master bean
     * @return bit length
     */
    protected int getWeightMatrixBitLength( ANN bean, double minW, double maxW ) {
        int retVal = 0;
        if ( bean != null ) {
            int size = 1 + bean.getIn() + bean.getHidden() + bean.getOut();
            double tmp = maxW - minW;
            if ( tmp > 0.0 )                
                retVal = getDoubleBitLength( tmp ) * size * size;
        }
        return retVal;
    }
    
    /**
     * Decode weight matrix from the chromosome.
     * <P>
     * @param chrom bit chromosome
     * @param index start index in the chromosome
     * @return new weight matrix
     */
    protected WeightMatrix decodeWeightMatrix( BitChromosome chrom, int index ) {
        WeightMatrix retVal = new WeightMatrix( bean.getIn(), bean.getHidden(), bean.getOut() );
        int size = 1 + bean.getIn() + bean.getHidden() + bean.getOut();
        double tmp = maxW - minW;
        int bs = getDoubleBitLength( tmp );
        int ci;
        double value;
        for ( int i = 0; i < size; i++ ) {
            for ( int j = 0; j < size; j++ ) {
                ci = index;
                index += bs;
                value = minW;
                value += decodeDouble( chrom.getBinaryInteger( ci, index - 1 ), tmp );
                if ( value > maxW )
                    value = maxW;
                else if ( value < minW )
                    value = minW;
                retVal.setWeight( i, j, value );
            }
        }
        return retVal;
    }
    
    /**
     * Get number of bits for encoding a double value.
     * It transforms the double to an integer, including the values after the comma.
     * Example: 3.5 -> 35 -> 100011, return value is 6
     * <P>
     * @param value double value
     * @return number of bits for this value
     */
    protected int getDoubleBitLength( double value ) {
        int retVal = 0;
        int commas = 1;
        if ( value == ( int ) value )
            commas = 2;
        double tmp = Math.pow( 10, ( new Double( value ) ).toString().length() + DOUBLE_CODING_EXT 
                     - ( new Integer( ( int ) value ) ).toString().length() - commas ) 
                     * value;
        retVal = Integer.toBinaryString( ( new Double( tmp ).intValue() ) ).length();
        return retVal;
    }
    
    /**
     * Decode double value from the binary integer value.
     * <P>
     * @param raw binary integer value
     * @param value original double value
     * @return decoded double value
     */
    protected double decodeDouble( int raw, double value ) {
        int commas = 1;
        if ( value == ( int ) value )
            commas = 2;
        return raw / Math.pow( 10, ( new Double( value ) ).toString().length() + DOUBLE_CODING_EXT
               - ( new Integer( ( int ) value ) ).toString().length() - commas );
    }
}
