package edu.cosy.netgen.ga;

import java.beans.*;
import banda.genalg.BitChromosome;

import edu.cosy.netgen.ann.*;

/**
 * Genotype for encoding a ANN bean with one chromosome.
 */
public class ANNGenotypeBasic extends ANNGenotype {
    
    /**
     * Copy constructor for this genotype.
     * <P>
     * @param genotype genotype to copy
     */
    public ANNGenotypeBasic( ANNGenotypeBasic genotype ) {
        super( genotype );
    }
    
    /**
     * Create a new genotype with one chromosome.
     * <P>
     * @param bean master ANN bean for encoding
     * @param co connection matrix encoding?
     * @param tp training paramteres encoding?
     * @param cy maximum training cycles encoding?
     * @param ac activation function encoding?
     * @param ma neuron markers encoding?
     * @param wt weight matrix encoding?
     * @param minW minimum value for weights
     * @param maxW maximum value for weights
     * @param afIn pool of activation functions for input neurons
     * @param afHidden pool of activation functions for hidden neurons
     * @param afOut pool of activation functions for output neurons
     */
    public ANNGenotypeBasic( ANN bean, boolean co, boolean tp, boolean cy, boolean ac,
                             boolean ma, boolean wt, double minW, double maxW, 
                             String[] afIn, String[] afHidden, String[] afOut ) {    
        super( 1 );       
        this.bean = bean;
        actFunctionPoolIn = afIn;
        actFunctionPoolHidden = afHidden;
        actFunctionPoolOut = afOut;
        int length = 0;
        if ( co ) {
            isCO = true;
            length += getConMatrixBitLength( bean );
        }
        if ( tp ) {
            isTP = true;
            length += getTrainingParamBitLength( bean );
        }
        if ( cy ) {
            isCY = true;
            length += getMaxTrainingCyclesBitLength( bean );
        }
        if ( ac ) {
            isAC = true;
            length += getActFunctionBitLength( bean );
        }
        if ( ma ) {
            isMA = true;
            length += getMarkersBitLength( bean );
        }
        if ( wt ) {
            isWT = true;
            this.minW = minW;
            this.maxW = maxW;
            length += getWeightMatrixBitLength( bean, minW, maxW );
        }
        setChromosome( 0, new BitChromosome( length ) );                
    }
    
    /**
     * Get the phenotype, an ANN bean, of this genotype.
     * <P>
     * @return decoded ANN bean, null if the genotype is emty
     */
    public Object getPhenotype() {
        ANN retVal = null;
        BitChromosome chrom = ( BitChromosome ) getChromosome();
        if ( bean != null && chrom.length() > 0 ) {
            try {
                retVal = ( ANN ) Beans.instantiate( null, "edu.cosy.netgen.ann.ANN" );
                retVal.setDescription( bean.getDescription() );
                retVal.setTrainingFunction( bean.getTrainingFunction() );
                retVal.setNetType( bean.getNetType() );
                retVal.setIn( bean.getIn() );
                retVal.setHidden( bean.getHidden() );
                retVal.setOut( bean.getOut() );
                retVal.setStopError( bean.getStopError() );

                int ci = 0;
                
                // set connection matrix
                if ( isCO )
                    retVal.setConMatrix( decodeConMatrix( chrom, ci ) ); 
                else
                    retVal.setConMatrix( new ConnectionMatrix( bean.getConMatrix() ) );         
                    
                // set training parameters
                if ( isTP )
                    retVal.setTrainingFunctionParams( decodeTrainingParams( chrom, ci ) );
                else {
                    if ( bean.getTrainingFunctionParams() != null ) {
                        TrainingParam[] tmp = new TrainingParam[ bean.getTrainingFunctionParams().length ];
                        System.arraycopy( bean.getTrainingFunctionParams(), 0, tmp, 0, tmp.length );
                        retVal.setTrainingFunctionParams( tmp );
                    }
                }
                
                // set training cycles
                if ( isCY ) {
                    retVal.setStopError( 0.0 );
                    retVal.setMaxTrainingCycles( decodeMaxTrainingCycles( chrom, ci ) );
                }
                else
                    retVal.setMaxTrainingCycles( bean.getMaxTrainingCycles() );
                    
                // set activation functions    
                if ( isAC )
                    retVal.setActivationFunctionsByNames( decodeActFunctions( chrom, ci ) );
                else {
                    if ( bean.getActivationFunctions() != null ) {
                        ActFunction[] tmp = new ActFunction[ bean.getActivationFunctions().length ];
                        System.arraycopy( bean.getActivationFunctions(), 0, tmp, 0, tmp.length );
                        retVal.setActivationFunctions( tmp );
                    }
                }
                
                // set neuron markers
                if ( isMA )
                    retVal.setNeuronMarkers( decodeMarkers( chrom, ci ) );
                else {
                    if ( bean.getNeuronMarkers() != null ) {
                        boolean[] tmp = new boolean[ bean.getNeuronMarkers().length ];
                        System.arraycopy( bean.getNeuronMarkers(), 0, tmp, 0, tmp.length );
                        retVal.setNeuronMarkers( tmp );
                    }
                }
                
                // set weight matrix
                if ( isWT )
                    retVal.setWeightMatrix( decodeWeightMatrix( chrom, ci ) );
                else
                    retVal.setWeightMatrix( new WeightMatrix( bean.getWeightMatrix() ) );
                
                retVal.setFitnessFunction( bean.getFitnessFunction() );
                retVal.setRegFunction( bean.getRegFunction() );
            } catch ( Exception e ) {
                return null;
            }
        }
        return retVal;
    }

    /**
     * Function for cloning.
     * <P>
     * @return new clone object
     */
    public Object clone() {
        return new ANNGenotypeBasic( this );
    }
}
