package edu.cosy.netgen.parser;

import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;
import fr.dyade.koala.xml.domlight.*;

import edu.cosy.netgen.ann.*;

/**
 * XML parser for ANN bean. This parser was validated with IBM XML-Parser xml4J_3_0_0EA3.
 */
public class AnnXMLParser implements AnnParser {
    
    // Description of the XML parser
    private final static String DESCRIPTION = "AnnML Format Parser for netGen 0.3";
    
    // extensions for net files
    private final static String[] NET_EXT = { "nml", "annml" };
    
    // extensions for pattern files
    private final static String[] PAT_EXT = { "pml", "patml" };
    
    /**
     * Get parser description.
     * <P>
     * @return description
     */
    public String getDescription() {
        return DESCRIPTION;
    }
    
    /**
     * Get net file extensions.
     * <P>
     * @return extensions
     */
    public String[] getNetFileExt() {
        return NET_EXT;
    }
    
    /**
     * Get pattern file extensions.
     * <P>
     * @return extensions
     */
    public String[] getPatternFileExt() {
        return PAT_EXT;
    }

    /**
     * Import net structure from an XML file.
     * <P>
     * @param file XML file
     * @return new bean instance
     */
    public ANN importNet( File file ) throws AnnParserException {
        ANN retVal = null;
        InputStream is = null;              
        try {
            is = new FileInputStream( file );
            XMLElementDeserializer deser = new XMLElementDeserializer( is );
            XMLElement root = deser.readXMLElement();
            deser.close();
            NetParams net = null;
            if ( root != null ) {
                net = new NetParams();

                XMLElement tmp = ( XMLElement ) root.getChild( 0 );
                net.date = ( ( XMLContent ) tmp.getChild( 0 ) ).getContent();                

                tmp = ( XMLElement ) root.getChild( 1 );
                net.desc = ( ( XMLContent ) tmp.getChild( 0 ) ).getContent();                

                tmp = ( XMLElement ) root.getChild( 2 );
                net.learnFunc = ( ( XMLContent ) tmp.getChild( 0 ) ).getContent();  
                
                tmp = ( XMLElement ) root.getChild( 3 );
                XMLElement unit = null;
                Unit tmpU = null;
                XMLElement tmpC = null;
                String tmpS = null;
                for ( int i = 0; i < tmp.getChildCount(); i++ ) {
                    unit = ( XMLElement ) tmp.getChild( i );
                    tmpU = new Unit();
                    tmpU.id = unit.getAttribute( "id" );

                    tmpC = ( XMLElement ) unit.getChild( 0 );
                    tmpU.nr = Integer.parseInt( ( ( XMLContent ) tmpC.getChild( 0 ) ).getContent() );

                    tmpC = ( XMLElement ) unit.getChild( 1 );
                    tmpU.bias = Double.parseDouble( ( ( XMLContent ) tmpC.getChild( 0 ) ).getContent() );

                    tmpC = ( XMLElement ) unit.getChild( 2 );
                    tmpS = ( ( XMLContent ) tmpC.getChild( 0 ) ).getContent();
                    if ( tmpS.equals( "i" ) )
                        tmpU.type = Unit.INPUT;
                    else if ( tmpS.equals( "h" ) )
                        tmpU.type = Unit.HIDDEN;
                    else if ( tmpS.equals( "o" ) )
                        tmpU.type = Unit.OUTPUT;

                    tmpC = ( XMLElement ) unit.getChild( 3 );
                    tmpU.actFunc = ( ( XMLContent ) tmpC.getChild( 0 ) ).getContent();

                    net.addUnit( tmpU );
                }

                net.init();
                tmp = ( XMLElement ) root.getChild( 4 );
                XMLElement connection = null;
                double w;
                for ( int i = 0; i < tmp.getChildCount(); i++ ) {
                    connection = ( XMLElement ) tmp.getChild( i );

                    tmpC = ( XMLElement ) connection.getChild( 0 );
                    w = Double.parseDouble( ( ( XMLContent ) tmpC.getChild( 0 ) ).getContent() );
                    
                    net.addConnection( connection.getAttribute( "src" ), connection.getAttribute( "dst" ), w );
                }                
            }

            // initialize bean
            retVal = ( ANN ) Beans.instantiate( null, "edu.cosy.netgen.ann.ANN" );            
            if ( retVal != null && net != null ) {
                retVal.setDescription( net.desc );
                retVal.setTrainingFunctionByName( net.learnFunc );
                retVal.setNetType( NeuralNet.FEEDFORWARD );
                retVal.setIn( net.in );
                retVal.setHidden( net.hidden );
                retVal.setOut( net.out );
                retVal.setConMatrix( net.cons );
                retVal.setWeightMatrix( net.weights );
                retVal.setActivationFunctionsByNames( net.actFunctions );
            }
        } catch( FileNotFoundException fnfe ) {
            throw new AnnParserException( "Net file " + file + " not found!" );
        } catch( Exception e ) {
            throw new AnnParserException( "Cannot read net file " + file + "!" + e );
        } finally {
            try {
                is.close();
            } catch ( IOException ioe ) {
            }
        }            
        return retVal;
    }
        
    /**
     * Export net structure into XML file.
     * <P>
     * @param file export file name
     * @param bean ANN instance
     * @return true, if success
     */
    public boolean exportNet( File file, ANN bean ) throws AnnParserException {
        boolean retVal = false;
        if ( bean != null ) {
            SimpleDateFormat dateFormatter = new SimpleDateFormat( "EEE MMM d H:mm:ss yyyy", Locale.US );
            DecimalFormatSymbols dfs = new DecimalFormatSymbols();
            dfs.setDecimalSeparator( '.' );
            DecimalFormat df = new DecimalFormat( "0.00000", dfs );
            
            String type = null;
            boolean[] markers = null;
            WeightMatrix weights = null;
            ConnectionMatrix cons = null;

            XMLElement root = new XMLElement( "ann" );
            XMLElement tmp = null;
            
            root.appendChild( tmp = new XMLElement( "date" ) );
            tmp.appendChild( new XMLContent( dateFormatter.format( new Date() ) ) );
            
            root.appendChild( tmp = new XMLElement( "desc" ) );
            tmp.appendChild( new XMLContent( XMLElementSerializer.convertStringToXML( bean.getDescription() ) ) );

            root.appendChild( tmp = new XMLElement( "learn_func" ) );
            tmp.appendChild( new XMLContent( bean.getTrainingFunctionByName() ) );
            
            root.appendChild( tmp = new XMLElement( "units" ) );
            markers = bean.getNeuronMarkers();
            weights = bean.getWeightMatrix();  
            XMLElement unit = null;
            XMLElement tmpE = null;
            for ( int i = 1; i <= bean.getIn() + bean.getHidden() + bean.getOut(); i++) {
                if ( markers == null || !markers[ i ] ) {
                    tmp.appendChild( unit = new XMLElement( "unit" ) );
                    if ( i <= bean.getIn() ) {
                        unit.addAttribute( "id", "i_" + i );
                        type = "i";
                    }
                    else if ( i <= bean.getIn() + bean.getHidden() ) {
                        unit.addAttribute( "id", "h_" + ( i - bean.getIn() ) );
                        type = "h";
                    }
                    else if ( i <= bean.getIn() + bean.getHidden() + bean.getOut() ) {
                        unit.addAttribute( "id", "o_" + ( i - ( bean.getIn() + bean.getHidden() ) ) );
                        type = "o";
                    }                   
                    
                    unit.appendChild( tmpE = new XMLElement( "nr" ) );
                    tmpE.appendChild( new XMLContent( Integer.toString( i ) ) );

                    unit.appendChild( tmpE = new XMLElement( "bias" ) );
                    tmpE.appendChild( new XMLContent( df.format( weights.getWeight( 0, i ) ) ) );
                    
                    unit.appendChild( tmpE = new XMLElement( "type" ) );
                    tmpE.appendChild( new XMLContent( type ) );                    

                    unit.appendChild( tmpE = new XMLElement( "act_func" ) );
                    tmpE.appendChild( new XMLContent( bean.getActivationFunctionsByNames()[ i ] ) );                    
                }
            }   
            
            root.appendChild( tmp = new XMLElement( "cons" ) );
            cons = bean.getConMatrix();
            XMLElement connection = null;
            for ( int i = 1; i <= cons.size(); i++ ) {
                for ( int j = 1; j <= cons.size(); j++ ) {
                    if ( cons.isConnection( i, j ) ) {
                        tmp.appendChild( connection = new XMLElement( "connection" ) );                            
                        if ( i <= bean.getIn() )
                            connection.addAttribute( "src", "i_" + i );
                        else if ( i <= bean.getIn() + bean.getHidden() )
                            connection.addAttribute( "src", "h_" + ( i - bean.getIn() ) );
                        else if ( i <= bean.getIn() + bean.getHidden() + bean.getOut() )
                            connection.addAttribute( "src", "o_" + ( i - ( bean.getIn() + bean.getHidden() ) ) );
    
                        if ( j <= bean.getIn() )
                            connection.addAttribute( "dst", "i_" + j );
                        else if ( j <= bean.getIn() + bean.getHidden() )
                            connection.addAttribute( "dst", "h_" + ( j - bean.getIn() ) );
                        else if ( j <= bean.getIn() + bean.getHidden() + bean.getOut() )
                            connection.addAttribute( "dst", "o_" + ( j - ( bean.getIn() + bean.getHidden() ) ) );
                            
                        connection.appendChild( tmpE = new XMLElement( "weight" ) );
                        tmpE.appendChild( new XMLContent( df.format( weights.getWeight( i, j ) ) ) );                            
                    }
                }
            }
            
            FileOutputStream out = null;
            try {
                out = new FileOutputStream( file );
                XMLElementSerializer ser = new XMLElementSerializer( out );
                ser.writeMetaData( "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" );
                ser.writeMetaData( "<!DOCTYPE ann SYSTEM \"net.dtd\">" );
                ser.writeXMLElement( root );
                ser.flush();
                ser.close();
            } catch( Exception e ) {
                throw new AnnParserException( "Cannot write net file " + file + "!" + e );
            } finally {
                try {
                    out.close();
                } catch ( IOException ioe ) {
                }
            }
        }
        return retVal;
    }    

    /**
     * Import patterns from XML file.
     * <P>
     * @param file pattern set file
     * @return pattern set
     */
    public PatternSet importPatterns( File file ) throws AnnParserException {
        PatternSet retVal = null;        
        InputStream is = null;
        try {
            is = new FileInputStream( file );
            XMLElementDeserializer deser = new XMLElementDeserializer( is );
            XMLElement root = deser.readXMLElement();
            deser.close();
            
            if ( root != null ) {
                retVal = new PatternSet();

                XMLElement tmp = ( XMLElement ) root.getChild( 1 );
                retVal.setDescription( ( ( XMLContent ) tmp.getChild( 0 ) ).getContent() );                

                XMLElement tmpE = null;
                String id = null;
                StringTokenizer tokenizer;
                double[] values = null;
                Vector in_pats = new Vector();
                Vector out_pats = new Vector();
                for ( int i = 2; i < root.getChildCount(); i++ ) {
                    tmp = ( XMLElement ) root.getChild( i );
                    id = tmp.getAttribute( "id" );
                    tmpE = ( XMLElement ) tmp.getChild( 0 );
                    tokenizer = new StringTokenizer( ( ( XMLContent ) tmpE.getChild( 0 ) ).getContent(), ";" );
                    values = new double[ tokenizer.countTokens() ];
                    int c = 0;                        
                    while ( tokenizer.hasMoreTokens() ) {
                        values[ c ] = Double.parseDouble( tokenizer.nextToken() );
                        c++;
                    }
                    if ( id.startsWith( "i" ) )
                        in_pats.add( new Pattern( values ) );
                    else if ( id.startsWith( "o" ) )
                        out_pats.add( new Pattern( values ) );
                }
                Pattern[] pset = new Pattern[ in_pats.size() ];
                in_pats.toArray( pset );
                retVal.setInPatterns( pset );
                pset = new Pattern[ out_pats.size() ];
                out_pats.toArray( pset );
                retVal.setOutPatterns( pset );
            }
            else
                retVal = null;
        } catch( FileNotFoundException fnfe ) {
            throw new AnnParserException( "Pattern file " + file + " not found!" );
        } catch( Exception e ) {
            throw new AnnParserException( "Cannot read pattern file " + file + "!" + e );
        } finally {
            try {
                is.close();
            } catch ( IOException ioe ) {
            }
        }
        return retVal;
    }    
    
    /**
     * Export patterns to XML file.
     * <P>
     * @param file pattern file
     * @param patterns pattern set
     * @return true, if success
     */
    public boolean exportPatterns( File file, PatternSet patterns ) throws AnnParserException {
        boolean retVal = false;
        if ( patterns != null ) {
            SimpleDateFormat dateFormatter = new SimpleDateFormat( "EEE MMM d H:mm:ss yyyy", Locale.US );
            DecimalFormatSymbols dfs = new DecimalFormatSymbols();
            dfs.setDecimalSeparator( '.' );
            DecimalFormat df = new DecimalFormat( "0.0", dfs );

            XMLElement root = new XMLElement( "set" );
            XMLElement tmp = null;

            root.appendChild( tmp = new XMLElement( "date" ) );
            tmp.appendChild( new XMLContent( dateFormatter.format( new Date() ) ) );
            
            root.appendChild( tmp = new XMLElement( "desc" ) );
            tmp.appendChild( new XMLContent( XMLElementSerializer.convertStringToXML( patterns.getDescription() ) ) );

            Pattern[] in_pats = patterns.getInPatterns();
            XMLElement pattern = null;
            XMLElement tmpE = null;
            StringBuffer tmpSB = null;
            for ( int i = 0; i < in_pats.length; i++ ) {
                root.appendChild( pattern = new XMLElement( "pattern" ) );
                pattern.addAttribute( "id", "i_" + ( i + 1 ) );
                pattern.appendChild( tmpE = new XMLElement( "values" ) );
                for ( int j = 0; j < in_pats[ i ].getSize(); j++ ) {
                    if ( tmpSB == null )
                        tmpSB = new StringBuffer();
                    tmpSB.append( Double.toString( in_pats[ i ].getValue( j ) ) );
                    if ( j < ( in_pats[ i ].getSize() - 1 ) )
                        tmpSB.append( ";" );
                }
                if ( tmpSB != null ) {
                    tmpE.appendChild( new XMLContent( tmpSB.toString() ) );
                    tmpSB = null;
                }
            }
            Pattern[] out_pats = patterns.getOutPatterns();
            for ( int i = 0; i < out_pats.length; i++ ) {
                root.appendChild( pattern = new XMLElement( "pattern" ) );
                pattern.addAttribute( "id", "o_" + ( i + 1 ) );
                pattern.appendChild( tmpE = new XMLElement( "values" ) );
                for ( int j = 0; j < out_pats[ i ].getSize(); j++ ) {
                    if ( tmpSB == null )
                        tmpSB = new StringBuffer();
                    tmpSB.append( Double.toString( out_pats[ i ].getValue( j ) ) );
                    if ( j < ( out_pats[ i ].getSize() - 1 ) )
                        tmpSB.append( ";" );
                }
                if ( tmpSB != null ) {
                    tmpE.appendChild( new XMLContent( tmpSB.toString() ) );
                    tmpSB = null;
                }
            }

            FileOutputStream out = null;
            try {
                out = new FileOutputStream( file );
                XMLElementSerializer ser = new XMLElementSerializer( out );
                ser.writeMetaData( "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" );
                ser.writeMetaData( "<!DOCTYPE set SYSTEM \"pat.dtd\">" );
                ser.writeXMLElement( root );
                ser.flush();
                ser.close();
                retVal = true;
            } catch( Exception e ) {
                throw new AnnParserException( "Cannot write net file " + file + "!" + e );
            } finally {
                try {
                    out.close();
                } catch ( IOException ioe ) {
                }
            }
        }
        return retVal;
   }    
}

class NetParams {
    public String date = null;
    public String desc = null;
    public String learnFunc = null;
    public Hashtable units = null;
    public ConnectionMatrix cons = null;
    public WeightMatrix weights = null;
    public boolean[] markers = null;
    public String[] actFunctions = null;
    
    public int in = 0;
    public int hidden = 0;
    public int out = 0;
    
    public void check() {
        if ( units != null ) {
        }           
    }
            
    public void addUnit( Unit u ) {
        if ( units == null )
            units = new Hashtable();
        units.put( u.id, u);
    }

    public void init() {
        if ( units != null && units.size() > 0 ) {
            actFunctions = new String[ units.size() + 1 ];
            Enumeration tmpUnits = units.elements();
            Unit tmpU = null;
            while ( tmpUnits.hasMoreElements() ) {
                tmpU = ( Unit ) tmpUnits.nextElement();
                if ( tmpU.type == Unit.INPUT )
                    in++;
                else if ( tmpU.type == Unit.HIDDEN )
                    hidden++;
                else 
                    out++;
                actFunctions[ tmpU.nr ] = tmpU.actFunc;
            }
            cons = new ConnectionMatrix( in, hidden, out, ConnectionMatrix.CREATE_BIAS_ONLY );
            weights = new WeightMatrix( in, hidden, out );
        }
    }
        
    public void addConnection( String s, String d, double w ) {
        Unit src = ( Unit ) units.get( s );
        Unit dest = ( Unit ) units.get( d );
        cons.setConnection( src.nr, dest.nr );
        weights.setWeight( 0, dest.nr, dest.bias );
        weights.setWeight( src.nr, dest.nr, w );
    }
    
    public String toString() {
        StringBuffer retVal = new StringBuffer("Date: ");
        retVal.append( date ).append( "\n" );
        retVal.append( "Description: " ).append( desc ).append( "\n" );
        retVal.append( "Learning function: " ).append( learnFunc ).append( "\n" );
        if ( units != null )
            retVal.append( "Units: " ).append( units.size() ).append( "\n" );
        if ( cons != null )
            retVal.append( "Connections: " ).append( cons.countConnections() );
        return retVal.toString();
    }
}

class Unit {
    public final static int INPUT = 0;
    public final static int HIDDEN = 1;
    public final static int OUTPUT = 2;
    
    public String id = null;
    public int nr = -1;
    public double bias = 0.0;
    public int type = -1;
    public String actFunc = null;
    
    public String toString() {
        return "(" + id + "), " + nr + ", " + bias + ", " + type + ", " + actFunc;
    }
}
