package edu.cosy.netgen.server;
 
import java.rmi.Naming;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;

import edu.cosy.netgen.client.service.ClientControlService;
import edu.cosy.netgen.server.event.ClientManagerEvent;
import edu.cosy.netgen.server.event.ClientManagerListener;
import edu.cosy.netgen.misc.PropertyHelper;

public class ClientManager {
    protected String key = "clientcmd_";
    protected Hashtable clientCmds = null;
    protected PropertyHelper properties = null;
    protected transient Vector listeners;
    protected Hashtable workingClients = null;
    protected Hashtable designClients = null;
    protected Hashtable managementClients = null;
    
    /**
     * Working client controll commands. The ';' is necessary, after that sign is the description of the command.
     * <P>
     * server.clientcmd_0 = start;Start working clients ( command for starting )
     * server.clientcmd_1 = stop;Stop working clients ( command for stopping )
     * server.clientcmd_2 = set;Set working object ( command for setting a working object )
     * server.clientcmd_3 = quit;Quit working clients ( command for exit the working client )
     * <P>
     */
    public ClientManager( PropertyHelper props, String key ) {
        properties = props;  
        if ( key != null )
            this.key = key;
        if ( properties != null ) {
            properties.addProperty( key + "0", ( new ClientControlCommand( 0, "start", "Start working clients" ) ).toString() );                        
            properties.addProperty( key + "1", ( new ClientControlCommand( 1, "stop", "Stop working clients" ) ).toString() );                        
            properties.addProperty( key + "2", ( new ClientControlCommand( 2, "set", "Set working object" ) ).toString() );                        
            properties.addProperty( key + "3", ( new ClientControlCommand( 3, "quit", "Quit working clients" ) ).toString() );
        }
    }
    
    public boolean initialize() {
        boolean retVal = false;
        if ( properties != null ) {
            clientCmds = new Hashtable();
            String[] keys = properties.getPropertyNames();
            ClientControlCommand tmp = null;
            try {
                for ( int i = 0; i < keys.length; i++ ) {
                    if ( keys[ i ].startsWith( "server.clientcmd_" ) ) {
                        tmp = ClientControlCommand.getCmdByString( properties.getProperty( keys[ i ] ) );
                        clientCmds.put( tmp.getCmd(), tmp );                        
                    }
                }
                retVal = true;
            } catch ( Exception e ) {
                return false;
            }
        }
        return retVal;
    }
    
    /**
     * Get the commands for controlling the working clients.
     * <P>
     * @return array of command strings
     */
    public ClientControlCommand[] getWorkingClientCommands() {
        ClientControlCommand[] retVal = null;
        if ( clientCmds != null ) {
            retVal = new ClientControlCommand[ clientCmds.size() ];
            retVal =  ( ClientControlCommand[] ) clientCmds.values().toArray( retVal );
        }
        return retVal;
    }

    public void close() {
    }
    
    public ClientControlCommand getCommand( String cmd ) {
        if ( clientCmds != null && cmd != null )
            return ( ClientControlCommand ) clientCmds.get( cmd );
        else
            return null;
    }
    
    public synchronized long connect( ClientInfo info ) {
        long retVal = -1;
        if ( info != null ) {
            switch ( info.getType() ) {
                case ClientInfo.DESIGN_CLIENT: {
                    retVal = connectDesignClient( ( DesignClientInfo ) info );
                    break;
                }
                case ClientInfo.MANAGEMENT_CLIENT: {
                    retVal = connectManagementClient( ( ManagementClientInfo ) info );
                    break;
                }
                case ClientInfo.WORKING_CLIENT: {
                    retVal = connectWorkingClient( ( WorkingClientInfo ) info );
                    break;
                }
                default: retVal = -1;
            }
        }
        if ( retVal != -1 )
            notifyClientConnectedEvent( new ClientManagerEvent( this, info ) );
        return retVal;
    }
    
    public synchronized boolean connect( long sid ) {
        boolean retVal = false;
        if ( workingClients != null ) {
            WorkingClientInfo info = ( WorkingClientInfo ) workingClients.get( new Long( sid ) );        
            if ( info != null && info.isAutoRunnable() ) {
                if ( connectWorkingClient( info ) == sid ) {
                    notifyClientConnectedEvent( new ClientManagerEvent( this, info ) );
                    retVal = true;
                }
            }
        }
        return retVal;
    }
    
    public synchronized boolean disconnect( long sid ) {
        boolean retVal = false;
        Long id = new Long( sid );
        ClientInfo info = null;
        if ( designClients != null && designClients.containsKey( id ) ) {
            info = ( DesignClientInfo ) designClients.get( id );
            retVal = disconnectDesignClient( ( DesignClientInfo ) info );
        } else if ( managementClients != null && managementClients.containsKey( id ) ) {
            info = ( ManagementClientInfo ) managementClients.get( id );
            retVal = disconnectManagementClient( ( ManagementClientInfo ) info );
        } else if ( workingClients != null && workingClients.containsKey( id ) ) {
            info = ( WorkingClientInfo ) workingClients.get( id );
            retVal = disconnectWorkingClient( ( WorkingClientInfo ) info );
        }
        if ( retVal )
            notifyClientDisconnectedEvent( new ClientManagerEvent( this, info ) );
        return retVal;
    }
    
    public synchronized boolean addWorkingClient( WorkingClientInfo info ) {
        boolean retVal = false;
        if ( workingClients == null)
            workingClients = new Hashtable();
        if ( info != null ) {
            workingClients.put(new Long( info.getSessionID() ), info);
            retVal = true;
        }
        return retVal;
    }

    public synchronized boolean removeWorkingClient( long id ) {
        boolean retVal = false;
        Long tmp = new Long( id );
        if ( workingClients.containsKey( tmp ) ) {
            workingClients.remove( tmp );
            retVal = true;
        }
        return retVal;
    }

    public int getWorkingClientCount() {
        if ( workingClients != null )
            return workingClients.size();
        else
            return 0;
    }
    
    public int getDesignClientCount() {
        if ( designClients != null )
            return designClients.size();
        else
            return 0;
    }
    
    public int startupWorkingClients() {
        int retVal = 0;
        WorkingClientInfo info = null;
        if ( workingClients != null ) {
            Enumeration e;
            synchronized( this ) {
                e = workingClients.elements();
            }
            for (; e.hasMoreElements(); ) {
                info = ( WorkingClientInfo ) e.nextElement();
                if ( info != null && info.isAutoRunnable() ) {
                    try {
                        Process process = Runtime.getRuntime().exec( info.getStartupSequence() );
                        if ( process != null ) {
                            retVal++;
                            notifyClientStartupEvent( new ClientManagerEvent( this, info ) );
                        }
                    } catch ( IOException ioe ) {
                    }
                }
            }
        }
        return retVal;
    }

    public ClientInfo getInfoFromID( int type, long sid ) {
        ClientInfo retVal = null;
        Long id = new Long( sid );
        switch ( type ) {            
            case ClientInfo.WORKING_CLIENT: {
                if ( workingClients != null && workingClients.containsKey( id ) )
                    retVal = ( WorkingClientInfo ) workingClients.get( id );
                break;
            }
            case ClientInfo.DESIGN_CLIENT: {
                if ( designClients != null && designClients.containsKey( id ) )
                    retVal = ( DesignClientInfo ) designClients.get( id );
                break;
            }
            case ClientInfo.MANAGEMENT_CLIENT: {
                if ( managementClients != null && managementClients.containsKey( id ) )
                    retVal = ( ManagementClientInfo ) managementClients.get( id );
                break;
            }
            default: retVal = null;
        }
        return retVal;
    }

    public int broadcastControlCommand( int type, ClientControlCommand cmd ) {
        int retVal = 0;
        if ( workingClients != null ) {
        switch ( type ) {            
            case ClientInfo.WORKING_CLIENT: {
                if ( workingClients != null ) {
                    Enumeration e;
                    WorkingClientInfo info = null;
                    synchronized( this ) {
                        e = workingClients.elements();
                    }
                    for (; e.hasMoreElements(); ) {
                        info = ( WorkingClientInfo ) e.nextElement();
                        if ( sendClientControlCommand( info, cmd ) )
                            retVal++;
                    }
                }
                break;
            }
            case ClientInfo.DESIGN_CLIENT: {
                break;
            }
            default: retVal = 0;
        }}
        return retVal;
    }
    
    public boolean sendClientControlCommand( ClientInfo info, ClientControlCommand cmd ) {
        boolean retVal = false;
        if ( info != null && info.isConnected() && info.getControlService() != null ) {
            try {
                info.getControlService().processCommand( cmd );
                retVal = true;
            } catch ( Exception e ) {
                retVal = false;
            }
        }
        return retVal;
    }
    
    private long connectDesignClient( DesignClientInfo info ) {
        long retVal = -1;
        if ( info != null ) {
            if ( designClients == null )
                designClients = new Hashtable();
            info.setSessionID( System.currentTimeMillis() );            
            designClients.put( new Long( info.getSessionID() ), info );
            info.setConnectionState( true );
            retVal = info.getSessionID();
        }
        return retVal;
    }
    
    private boolean disconnectDesignClient( DesignClientInfo info ) {
        if ( info != null ) {
            designClients.remove( new Long( info.getSessionID() ) );
            return true;
        } else
            return false;
    }
    
    private long connectManagementClient( ManagementClientInfo info ) {
        long retVal = -1;
        if ( info != null ) {
            if ( managementClients == null )
                managementClients = new Hashtable();
            info.setSessionID( System.currentTimeMillis() );            
            managementClients.put( new Long( info.getSessionID() ), info );
            info.setConnectionState( true );
            retVal = info.getSessionID();
        }
        return retVal;
    }
    
    private boolean disconnectManagementClient( ManagementClientInfo info ) {
        if ( info != null ) {
            managementClients.remove( new Long( info.getSessionID() ) );
            return true;
        } else
            return false;
    }

    private long connectWorkingClient( WorkingClientInfo info ) {
        long retVal = -1;
        if ( info != null ) {
            try {
                ClientControlService cs = ( ClientControlService ) Naming.lookup( "rmi://" + info.getAddress().getHostAddress() + ":" + info.getPort() + "/clientControlService" ); 
                info.setControlService( cs );
            } catch ( Exception e ) {
                return -1;
            }
            if ( !info.isAutoRunnable() ) {
                info.setSessionID( System.currentTimeMillis() );
                addWorkingClient( info );
            }
            info.setConnectionState( true );
            retVal = info.getSessionID();
        }
        return retVal;
    }
    
    private boolean disconnectWorkingClient( WorkingClientInfo info ) {
        boolean retVal = false;
        if ( info != null ) {
            info.setConnectionState( false );
            info.setControlService( null );
            if ( !info.isAutoRunnable() )
                removeWorkingClient( info.getSessionID() );
            retVal = true;             
        }
        return retVal;
    }

    public WorkingClientInfo[] getWorkingClientList( boolean auto ) {
        WorkingClientInfo[] retVal = null;
        WorkingClientInfo[] tmp = ( WorkingClientInfo[] ) getClientList( ClientInfo.WORKING_CLIENT );
        if ( tmp != null && tmp.length > 0 ) {
            if ( auto ) {
                Vector list = new Vector();
                for ( int i = 0; i < tmp.length; i++ ) {
                    WorkingClientInfo info = ( WorkingClientInfo ) tmp[ i ];                    
                    if ( info.isAutoRunnable() )
                        list.add( info );
                }
                retVal = ( WorkingClientInfo[] ) list.toArray( retVal );
            }
            else
                retVal = tmp;
        }
        return retVal;
    }
    
    public DesignClientInfo[] getDesignClientList() {
        return ( DesignClientInfo[] ) getClientList( ClientInfo.DESIGN_CLIENT );
    }
    
    public ManagementClientInfo[] getManagementClientList() {
        return ( ManagementClientInfo[] ) getClientList( ClientInfo.MANAGEMENT_CLIENT );
    }

    private ClientInfo[] getClientList( int type ) {
        ClientInfo[] retVal = null;
        Hashtable tmp = null;
        switch ( type ) {
            case ClientInfo.WORKING_CLIENT: {
                tmp = workingClients;
                retVal = new WorkingClientInfo[ 0 ];
                break;
            }
            case ClientInfo.DESIGN_CLIENT: {
                tmp = designClients;
                retVal = new DesignClientInfo[ 0 ];
                break;
            }
            case ClientInfo.MANAGEMENT_CLIENT: {
                tmp = managementClients;
                retVal = new ManagementClientInfo[ 0 ];
                break;
            }
        }
        if ( tmp != null && tmp.size() > 0 ) {
            Enumeration entries = tmp.elements();
            ClientInfo tmpInfo = null;
            Vector list = new Vector();
            while ( entries.hasMoreElements() ) {
                tmpInfo = ( ClientInfo ) entries.nextElement();
                list.add( tmpInfo );   
            }                
            retVal = ( ClientInfo[] ) list.toArray( retVal );
        }
        return retVal;
    }

    public synchronized void addListener( ClientManagerListener l ) {
        if ( listeners == null )
            listeners = new Vector();
        listeners.add( l );
    }
    
    public synchronized void removeListener( ClientManagerListener l ) {
        if ( listeners != null && l != null )
            listeners.remove( l );
    }
    
    public synchronized ClientManagerListener[] getListeners() {
        Vector list;
        ClientManagerListener[] retVal = null;
        if ( listeners != null ) {
            synchronized ( this ) {
                list = ( Vector ) listeners.clone();
            }
            retVal = ( ClientManagerListener[] ) list.toArray();
        }
        return retVal;
    }    

    public void notifyClientConnectedEvent( ClientManagerEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                ClientManagerListener listener = ( ClientManagerListener ) list.elementAt( i );
                listener.clientConnected( e );
            }
        }
    }
    
    public void notifyClientDisconnectedEvent( ClientManagerEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                ClientManagerListener listener = ( ClientManagerListener ) list.elementAt( i );
                listener.clientDisconnected( e );
            }
        }
    }

    public void notifyClientStartupEvent( ClientManagerEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                ClientManagerListener listener = ( ClientManagerListener ) list.elementAt( i );
                listener.clientStartup( e );
            }
        }
    }
}
