package edu.cosy.netgen.server;

import java.util.Vector;

import edu.cosy.netgen.server.event.TaskListener;
import edu.cosy.netgen.server.event.TaskEvent;
import edu.cosy.netgen.server.event.FIFOListener;
import edu.cosy.netgen.server.event.FIFOEvent;

/**
 * Implementation of FIFO queue. 
 * If there is a task to process and a working client is connected, then it starts working.
 */
public class FIFO implements Runnable, TaskListener {
    
    // listeners for FIFO events
    private Vector listeners;
    
    // the queue
    private Vector queue;
    
    // flag if one task is running
    private boolean isOneRunning = true;
    
    // handle for current task
    private Task currentTask;
    
    // worker thread
    private Thread worker;
    
    // flag for suspending
    private boolean isSuspended = true;
    
    /**
     * Default constructor.
     */    
    public FIFO() {
        super();
    }

    /**
     * Suspend the queue.
     */
    public synchronized void suspendWorking() {
        isSuspended = true;
        notifyAll();
    }
    
    /**
     * Resume the queue.
     */
    public synchronized void resumeWorking() {
        isSuspended = false;
        isOneRunning = false;
        notifyAll();
    }
    
    /**
     * Add a new task to the queue.
     * <P>
     * @param t task to be added
     */
    public synchronized void add( Task t ) {
        if ( t != null ) {
            if ( queue == null )
                queue = new Vector();
            t.getRuntimeInfo().setState( TaskRuntimeInfo.QUEUED );
            queue.add( t );
            notifyAll();
        }
    }
      
    /**
     * Remove a task from the queue. If its the current running task, the queue is interrupted.
     * <P>
     * @param t task to be removed
     * @return true, if the task could be removed
     */
    public synchronized boolean remove( Task t ) {
        boolean retVal = false;
        if ( queue != null && t != null ) {
            if ( t == currentTask ) {
                suspendWorking();
                isOneRunning = false;
            }
            if ( queue.contains( t ) ) {
                queue.remove( t );      
                retVal = true;
            }     
            if ( isSuspended )
                resumeWorking();     
        }
        return retVal;
    }
    
    /**
     * Get the information if a task is stored in the FIFO queue.
     * <P>
     * @param t task to query
     * @return true, if task is in the queue
     */   
    public synchronized boolean contains( Task t ) {
        if ( queue != null && t != null && queue.contains( t ) )
            return true;
        else
            return false;
    }
       
    /**
     * Working part of the FIFO queue. Fires events if it waits or starts working.
     */ 
    public void run() {
        while ( true ) {
            try {
                synchronized ( this ) {
                    while ( queue == null || queue.size() == 0 || isOneRunning || isSuspended ) {
                        notifyWaitingEvent( new FIFOEvent( this ) );
                        wait();    
                    }
                }
            } catch ( InterruptedException ie ) {
                return;
            }
            currentTask = ( Task ) queue.get( 0 );
            currentTask.addListener( this );
            notifyStartWorkingEvent( new FIFOEvent( this, currentTask ) );
            worker = new Thread( currentTask, "FIFOWorker" );
            worker.start();
            isOneRunning = true;
        }
    }
    
    /**
     * Add a listener for FIFO events.
     * <P>
     * @param l listener to be added
     */
    public synchronized void addListener( FIFOListener l ) {
        if ( listeners == null )
            listeners = new Vector();
        listeners.add( l );
    }
    
    /**
     * Remove a FIFO listener.
     * <P>
     * @param l listener to be removed.
     */
    public synchronized void removeListener( FIFOListener l ) {
        if ( listeners != null && l != null )
            listeners.remove( l );
    }
    
    /**
     * Get all lsiteners for this FIFO queue.
     * <P>
     * @return array of FIFO listeners
     */
    public synchronized FIFOListener[] getListeners() {
        Vector list;
        FIFOListener[] retVal = null;
        if ( listeners != null ) {
            synchronized ( this ) {
                list = ( Vector ) listeners.clone();
            }
            retVal = ( FIFOListener[] ) list.toArray();
        }
        return retVal;
    }    
    
    /**
     * Function implemented for TaskListener.
     * <P>
     * @param e task event
     */
    public void taskInitialized( TaskEvent e ) {
    }
    
    /**
     * Function implemented for TaskListener.
     * <P>
     * @param e task event
     */
    public void taskStartWorking( TaskEvent e ) {
    }
    
    /**
     * Function implemented for TaskListener.
     * <P>
     * @param e task event
     */
    public void taskStopWorking( TaskEvent e ) {
    }
    
    /**
     * Function implemented for TaskListener. The task is removed from the queue and a event is fired.
     * If there is one more task in queue, it starts working again.
     * <P>
     * @param e task event
     */
    public synchronized void taskReady( TaskEvent e ) {
        if ( e != null && e.getSource() != null ) {
            currentTask.removeListener( this );
            worker = null;
            queue.remove( currentTask );
            isOneRunning = false;
            notifyStopWorkingEvent( new FIFOEvent( this, ( Task ) e.getSource() ) );
            notifyAll();
        }
    }
    
    // notify that the queue is started
    private void notifyStartWorkingEvent( FIFOEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                FIFOListener listener = ( FIFOListener ) list.elementAt( i );
                listener.fifoStartWorking( e );
            }
        }
    }

    // notify that the queue is waiting
    private void notifyWaitingEvent( FIFOEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                FIFOListener listener = ( FIFOListener ) list.elementAt( i );
                listener.fifoWaiting( e );
            }
        }
    }

    // notify that the queue stops working
    private void notifyStopWorkingEvent( FIFOEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                FIFOListener listener = ( FIFOListener ) list.elementAt( i );
                listener.fifoStopWorking( e );
            }
        }
    }
}
