package edu.cosy.netgen.server;

import java.io.*;
import java.net.*;
import java.util.*;
import java.rmi.*;
import java.rmi.server.*;
import java.rmi.registry.LocateRegistry;
import java.beans.Beans;
import org.log4j.FileAppender;
import org.log4j.PatternLayout;
import org.log4j.Category;
import org.log4j.PropertyConfigurator;

import edu.cosy.netgen.server.service.*;
import edu.cosy.netgen.client.service.ClientControlService;
import edu.cosy.netgen.server.event.FIFOEvent;
import edu.cosy.netgen.server.event.FIFOListener;
import edu.cosy.netgen.server.event.ContainerEvent;
import edu.cosy.netgen.server.event.ContainerListener;
import edu.cosy.netgen.server.event.ClientManagerEvent;
import edu.cosy.netgen.server.event.ClientManagerListener;
import edu.cosy.netgen.misc.PropertyHelper;
import edu.cosy.netgen.misc.Constants;

/**
 * The netGen 0.3 Server. It distributes the working and coordinates the working and design clients.
 * All working is stored in a repository. If something goes wrong, while a task is processed,
 * it restarts working than the server is started again.
 */
public class NetGenServer implements FIFOListener, ContainerListener, ClientManagerListener {
    static Category CAT = Category.getInstance( "Server" );    

    // filename of the properties
    private String SERVER_PROPERTIES = "server.properties";
    private PrintWriter logWriter;
    
    // properties   
    private PropertyHelper properties;
    
    // client service instance
    private DesignClientServiceImpl clientService;
    
    // control service instance
    private ServerControlServiceImpl controlService;
        
    // container service instance
    private WorkingClientServiceImpl containerService;
    
    private ClientManager clientManager;
    
    // list of all tasks    
    private Hashtable taskContainer;
    
    // the repository
    private Repository repository;   
    
    // the worker thread
    private Thread worker;
    
    // the FIFO queue
    private FIFO fifo;
    
    // the container for distributing the individuals to be calculated
    private Container container;
    
    // flag for running state
    private boolean isRunning = false;
    
    /**
     * Create a new server instance. The property file is loaded and the server is initialized.
     */
    public NetGenServer() {
        super();
        properties = new PropertyHelper( SERVER_PROPERTIES, "netGen " + Constants.VERSION + " Properties" );
        if ( !initialize() )
            exit( 1 );
    }
    
    /**
     * Initialize the server. If there are stored tasks in the repository, they are attached to the new FIFO queue.
     * Also the list with the working clients for automatic startup is loaded. The following properties are set:
     * <P>
     * server.rmi.host    = localhost                           ( host there the server is running )
     * server.rmi.port    = 12340                               ( port there the server is publishing its services )
     * server.log.dir     = logs/                               ( directory where the log file is stored )
     * server.log.file    = netGen.log                          ( log file name )     
     * server.log.format  = %-6r [%15.15t] %-5p %30.30c %x - %m ( log file format, look at the documentation of log4j for more information )     
     * <P>
     * @return true, if all goes well
     */
     private boolean initialize() {
        boolean retVal = false;        
        if ( properties != null && properties.open() ) {        
            properties.addProperty( "server.rmi.host", "localhost" );
            properties.addProperty( "server.rmi.port", "12340" ); 
            properties.addProperty( "server.log.dir", "log/" );             
            properties.addProperty( "server.log.file", "netGen.log" );   
            properties.addProperty( "server.log.format", "%-6r [%15.15t] %-5p %30.30c %x - %m" );
            if ( initLog() ) {
                CAT.info( "Initializing netGen Server ..." );
                repository = Repository.createRepository( properties, "edu.cosy.netgen.server.DiskRepository" );
                taskContainer = new Hashtable();
                fifo = new FIFO();
                fifo.addListener( this );
                container = new Container();   
                container.addListener( this );   
                clientManager = new ClientManager( properties, "server.clientcmd_" );
                clientManager.addListener( this );
                if ( initializeRMI() && repository.initRepository() && clientManager.initialize() ) {
                    
                    // loading tasks from repository asynchron
                    final long[] ids = repository.listStoredTasks();                    
                    Thread taskLoader = new Thread() {
                        public void run() {
                            if ( ids != null ) {    
                                Task tmp = null;
                                for ( int i = 0; i < ids.length; i++ ) {
                                    if ( repository != null && ( tmp = repository.loadTask( ids[ i ] ) ) != null ) {
                                        taskContainer.put( new Long( ids[ i ] ), tmp );
                                        if ( tmp.getRuntimeInfo().getState() != TaskRuntimeInfo.READY ) {
                                            tmp.setContainer( container );
                                            fifo.add( tmp );
                                            CAT.info( "Task " + ids[ i ] + " added to FIFO queue!" );
                                        } else
                                            CAT.info( "Task " + ids[ i ] + " loaded from repository!" );
                                    } else
                                        CAT.warn( "Cannot add task " + ids[ i ] + " to task container!" );
                                }
                            }
                        }
                    };
                    taskLoader.start();
                    
                    // loading working client list
                    Thread workerLoader = new Thread() {
                        public void run() {
                            WorkingClientInfo[] wcList = repository.loadWorkingClientList();
                            if ( wcList != null && wcList.length > 0 )
                                for ( int i = 0; i < wcList.length; i++ ) {
                                    clientManager.addWorkingClient( wcList[ i ] );
                                    CAT.info( "Working client " + wcList[ i ].getAddress() + " loaded from repository!" );
                                }
                        }
                    };
                    workerLoader.start();
                            
                    retVal = true;
                }
            }
        }
        return retVal;
    }
    
    /**
     * Start netGen server.
     */
    public synchronized void start() {
        CAT.info( "Starting netGen Server ... " );
        if ( worker == null )
            worker = new Thread( fifo );
        worker.start();
        isRunning = true;
    }
    
    /**
     * Stop netGen server.
     */
    public synchronized void stop() {
        worker = null;
        CAT.info( "Stopping netGen Server ... " );
        isRunning = false;
    }
    
    // Initialize RMI and bind objects to naming service
    private boolean initializeRMI() {
        boolean retVal = false;
        try {
            CAT.info( "Starting RMI server ... " );
                            
            // create registry service
            LocateRegistry.createRegistry( Integer.parseInt( properties.getProperty( "server.rmi.port" ) ) );
            
            // instanciate client service
            clientService = new DesignClientServiceImpl( this, clientManager );
            
            // bind object
            retVal = bindRMIObject( clientService, "netGenClientService", true );
            
            // instanciate control service
            controlService = new ServerControlServiceImpl( this, clientManager );
            
            // bind object
            retVal = bindRMIObject( controlService, "netGenControlService", true );
            
            // instanciate container service
            containerService = new WorkingClientServiceImpl( clientManager, container );
            
            // bind object
            retVal = bindRMIObject( containerService, "netGenContainerService", true );
        } catch ( Exception e ) {
            CAT.error( "Cannot initialize RMI server! " + e );
            retVal = false;
        }
        return retVal;
    }

    // bind/unbind an RMI object
    private boolean bindRMIObject( Remote obj, String name, boolean flag ) throws RemoteException, MalformedURLException, NotBoundException {
        boolean retVal = false;
        String rmiHost = properties.getProperty( "server.rmi.host" );
        String rmiPort = properties.getProperty( "server.rmi.port" );
        if ( flag ) {
            Naming.rebind( "rmi://" + rmiHost + ":" + rmiPort + "/" + name, obj);
            CAT.info( name + "\tbound" );   
            retVal = true;
        } else {
            Naming.unbind( "rmi://" + rmiHost + ":" + rmiPort + "/" + name);
            CAT.info( name + "\tunbound" ); 
            retVal = true;
        }
        return retVal;
    }
            
    /**
     * Create a new task. It is stored in the repository.
     * <P>
     * @param info task creation information
     * @return new task ID
     */
    public synchronized long createTask( TaskCreationInfo info ) {
        long retVal = -1;        
        if ( info != null && container != null ) {
            long id = System.currentTimeMillis();
            Task task = new Task( id, container, info );
            if ( task != null ) {
                if ( fifo != null && taskContainer != null ) {
                    taskContainer.put( new Long( id ), task );
                    fifo.add( task );
                    if ( repository.storeTask( task ) ) 
                        retVal = id;
                }
            }
        }
        return retVal;
    }
    
    /**
     * Get the runtime information of an task.
     * <P>
     * @param taskID task ID
     * @return task runtime information
     */
    public synchronized TaskRuntimeInfo getTaskInformation( long taskID ) {
        TaskRuntimeInfo retVal = null;
        Task task = ( Task ) taskContainer.get( new Long( taskID ) );
        if ( task != null)
            retVal = task.getRuntimeInfo();
        return retVal;
    }
    
    /**
     * Get information about all tasks on the server.
     * <P>
     * @return array of task runtime information
     */
    public synchronized TaskRuntimeInfo[] getTaskList() {
        TaskRuntimeInfo[] retVal = null;
        if ( taskContainer != null )
            retVal = ( TaskRuntimeInfo[] ) taskContainer.values().toArray( retVal );
        return retVal;
    }
    
    /**
     * Get the result of a processed task.
     * <P>
     * @param taskID task ID
     * @param remove if true, than the task is removed from the repository
     * @return result of the task, null if there is no result
     */
    public synchronized TaskResult getTaskResult( long taskID, boolean remove ) {
        TaskResult retVal = null;
        Long id = new Long( taskID );
        if ( taskContainer.containsKey( id ) ) {
            Task task = ( Task ) taskContainer.get( id );
            if ( task != null && task.getRuntimeInfo().getState() == TaskRuntimeInfo.READY )
                retVal = task.getResult();
            if ( retVal != null && repository.isTaskStored( taskID ) && remove ) {
                taskContainer.remove( id );
                repository.deleteStoredTask( taskID );
            }
        }
        return retVal;
    }
    
    /**
     * Remove task from server.
     * <P>
     * @param taskID task ID for the task to be removed
     * @param remove true, if the task should be removed from the repository
     * @return true, if the task was removed
     */
    public synchronized boolean removeTask( long taskID, boolean remove ) {
        boolean retVal = false;
        Long id = new Long( taskID );
        if ( taskContainer.containsKey( id ) ) {
            Task old = ( Task ) taskContainer.get( id );
            if ( fifo.contains( old ) )
                retVal = fifo.remove( old );
            if ( remove && repository.isTaskStored( taskID ) )
                retVal = repository.deleteStoredTask( taskID ); 
            if ( retVal = true )
                taskContainer.remove( id );
        }
        return retVal;
    }
    
    /**
     * Get the handle of the distribution container.
     * <P>
     * @return container handle
     */
    public Container getContainer() {
        return container;
    }
    
    /**
     * Check if the server is running.
     * <P>
     * @return true, if server is running
     */
    public boolean isRunning() {
        return isRunning;
    }
    
    /**
     * Close the server and shut down all services. 
     * All tasks and the list of working clients with capable for automatic startup are stored in the repository.
     * All connected working clients are also be shut down.
     * <P>
     * @param state exit state
     */
    public void exit( int state ) {
        CAT.info( "Shut down netGen server ... " );
        if ( state == 1)
            CAT.warn( "Abnormal termination" );
        stop();
        clientManager.broadcastControlCommand( ClientInfo.WORKING_CLIENT, clientManager.getCommand( "quit" ) );
        if ( repository != null ) {
            repository.storeWorkingClientList( clientManager.getWorkingClientList( true ) );
            repository.closeRepository();
        }
        System.out.println( "Shut down RMI server ... " );      
        try {
            bindRMIObject( null, "netGenClientService", false );
            bindRMIObject( null, "netGenControlService", false );
            bindRMIObject( null, "netGenContainerService", false );
        } catch ( Exception e ) {
            CAT.error( "Cannot unbind RMI objects! " + e );
        }
        if ( properties != null )
            properties.close();
        if ( logWriter != null )
            logWriter.close();
        System.exit( state );
    }
    
    /**
     * React on FIFO event.
     * <P>
     * @param e FIFO event
     */    
    public void fifoStartWorking( FIFOEvent e ) {
        if ( e != null && e.getCurrentTask() != null ) {
            ClientControlCommand cmd = clientManager.getCommand( "set" );
            cmd.setParam( e.getCurrentTask().getWorkingObject() );
            clientManager.broadcastControlCommand( ClientInfo.WORKING_CLIENT, cmd );
        }
    }
    
    /**
     * React on FIFO event.
     * <P>
     * @param e FIFO event
     */    
    public void fifoWaiting( FIFOEvent e ) {
    }
    
    /**
     * React on FIFO event.
     * <P>
     * @param e FIFO event
     */    
    public void fifoStopWorking( FIFOEvent e ) {
        if ( e != null ) {
            Task task = e.getCurrentTask();
            if ( task != null )
                repository.storeTask( task );
        }
    }
    
    /**
     * React on container event.
     * <P>
     * @param e container event
     */    
    public void containerInitialized( ContainerEvent e ) {
        if ( e != null ) 
            clientManager.broadcastControlCommand( ClientInfo.WORKING_CLIENT, clientManager.getCommand( "start" ) );
    }
    
    /**
     * React on container event.
     * <P>
     * @param e container event
     */    
    public void containerCalculated( ContainerEvent e ) {
        if ( e != null ) 
            clientManager.broadcastControlCommand( ClientInfo.WORKING_CLIENT, clientManager.getCommand( "stop" ) );
    }
    
    /**
     * React on container event.
     * <P>
     * @param e container event
     */    
    public void containerIndividualDistributed( ContainerEvent e ) {
    }    
    
    public void clientConnected( ClientManagerEvent e ) {
        if ( fifo != null )
            fifo.resumeWorking();
    }
    
    public void clientDisconnected( ClientManagerEvent e ) {
        if ( fifo != null && clientManager.getWorkingClientCount() == 0 )
            fifo.suspendWorking();
    }

    public void clientStartup( ClientManagerEvent e ) {
    }
    
    /**
     * Main function for startup. The command "quit" shut down the server.
     * <P>
     * @param args parameters for startup
     */
    public static void main( String[] args ) {
        NetGenServer main = new NetGenServer();
        if ( main == null ) {
            System.err.println( "Cannot start netGen Server!" );
            System.exit( 1 );
        }
        else {
            main.start();
            System.out.println( "netGen Server is working..." );
            System.out.print( "Type \"quit\" for exit ..." );
            try {
                byte[] command = new byte[ 4 ];
                System.in.read(command, 0, 4);
                if ( ( new String( command ) ).equals( "quit" ) )
                    main.exit( 0 );
            }
            catch (Exception e) {
            }
        }
    }
        
    private boolean initLog() {
        PatternLayout layout = new PatternLayout( properties.getProperty( "server.log.format" ) + "\n" );
        try {
            File logDir = new File( properties.getProperty( "server.log.dir" ) );
            if ( !logDir.exists() )
                logDir.mkdir();
            File logFile = new File( logDir, properties.getProperty( "server.log.file" ) );
            if ( !logFile.exists() )
                logFile.createNewFile();
            logWriter = new PrintWriter( new FileOutputStream( logFile.toString() ) );
            PropertyConfigurator.configure( new FileAppender( layout, logWriter ) );
        } catch ( Exception e ) {
            System.err.println( "Cannot open log file! " + e );
            return false;
        }
        return true;
    }
}
