package edu.cosy.netgen.server;
 
import java.util.Vector;

import banda.genalg.GeneticAlgorithm;
import banda.genalg.Population;
import banda.genalg.Evaluation;
    
import edu.cosy.netgen.server.event.ContainerEvent;
import edu.cosy.netgen.server.event.ContainerListener;
import edu.cosy.netgen.server.event.TaskListener;
import edu.cosy.netgen.server.event.TaskEvent;
import edu.cosy.netgen.ga.ANNGenotype;

public class Task implements Runnable, ContainerListener, java.io.Serializable {
    private transient Vector listeners;
    private long taskID = -1;
    private TaskCreationInfo creationInfo;
    private transient Container container;
    private TaskResult result;
    
    private boolean isSuspended = true;
    private TaskRuntimeInfo runtimeInfo;
        
    public Task( long id, Container c, TaskCreationInfo ci ) {
        taskID = id;
        container = c;
        creationInfo = ci;
        result = new TaskResult( this );
        runtimeInfo = new TaskRuntimeInfo( this );
    }

    public void setContainer( Container c ) {
        container = c;
    }
    
    public void run() {
        if ( container != null && creationInfo != null ) {
            runtimeInfo.setStartTime( System.currentTimeMillis() );
            container.addListener( this );
            Thread f = Thread.currentThread();
            GeneticAlgorithm ga = creationInfo.createGeneticAlgorithm();
            try {
                ga.initialize( creationInfo.getPopulationSize() );           
            } catch ( Exception e ) {
                return;
            }
            notifyInitializedEvent( new TaskEvent( this ) );
            for ( int i = 0; i < creationInfo.getGenerations() + 1; i++ ) {
                notifyStartWorkingEvent( new TaskEvent( this ) );
                container.initialize( ga.getPopulation() );
                container.setCurrentTaskType( creationInfo.getTaskType() );
                try {
                    synchronized ( this ) {
                        while ( isSuspended )
                            wait();
                    }
                } catch ( InterruptedException ie ) {
                }
                result.addPopulation( i, ga.getPopulation() );
                ga.evolve();
                isSuspended = true;
                runtimeInfo.setPercentage( ( ( i + 1 ) / creationInfo.getGenerations() ) * 100 );
                notifyStopWorkingEvent( new TaskEvent( this ) );
            }
            
//if ( result != null )
//    System.out.println( "F: " + result.getBestGenotype().getFitness() );            
    
            container.removeListener( this );
            runtimeInfo.setStopTime( System.currentTimeMillis() );
            notifyReadyEvent( new TaskEvent( this ) );
        }
    }
        
    public synchronized TaskResult getResult() {
        return result;
    }

    public long getTaskID() {
        return taskID;
    }
        
    public TaskRuntimeInfo getRuntimeInfo() {
        return runtimeInfo;
    }

    public java.io.Serializable getWorkingObject() {
        java.io.Serializable retVal = null;
        if ( creationInfo != null ) {
            switch ( creationInfo.getTaskType() ) {
                case TaskCreationInfo.EVOLVE_ANN: {
                    retVal = ( java.io.Serializable ) creationInfo.getWorkingPatternSets();
                    break;
                }
                case TaskCreationInfo.EVOLVE_PATTERN: {
                    Vector tmp = new Vector();
                    tmp.add( creationInfo.getBean() );
                    tmp.add( creationInfo.getWorkingPatternSets() );
                    retVal = ( java.io.Serializable ) tmp;
                }
            }
        }
        return retVal;
    }
        
    public synchronized void containerCalculated( ContainerEvent e ) {
        if ( e != null && ( Container ) e.getSource() == container ) {
            isSuspended = false;
            notifyAll();
        }
    }
    
    public void containerInitialized( ContainerEvent e ) {
    }
    
    public void containerIndividualDistributed( ContainerEvent e ) {
    }
    
    public synchronized void addListener( TaskListener l ) {
        if ( listeners == null )
            listeners = new Vector();
        listeners.add( l );
    }
    
    public synchronized void removeListener( TaskListener l ) {
        if ( listeners != null && l != null )
            listeners.remove( l );
    }
    
    public synchronized TaskListener[] getListeners() {
        Vector list;
        TaskListener[] retVal = null;
        if ( listeners != null ) {
            synchronized ( this ) {
                list = ( Vector ) listeners.clone();
            }
            retVal = ( TaskListener[] ) list.toArray();
        }
        return retVal;
    }    
            
    public void notifyInitializedEvent( TaskEvent e) {
        Vector list;
        synchronized( this ) {
            list = ( Vector ) listeners.clone();
        }
        for ( int i = 0; i < list.size(); i++ ) {
            TaskListener listener = ( TaskListener ) list.elementAt( i );
            listener.taskInitialized( e );
        }
    }

    public void notifyReadyEvent( TaskEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                TaskListener listener = ( TaskListener ) list.elementAt( i );
                listener.taskReady( e );
            }
        }
    }

    public void notifyStartWorkingEvent( TaskEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                TaskListener listener = ( TaskListener ) list.elementAt( i );
                listener.taskStartWorking( e );
            }
        }
    }
    
    public void notifyStopWorkingEvent( TaskEvent e) {
        Vector list;
        if ( listeners != null ) {
            synchronized( this ) {
                list = ( Vector ) listeners.clone();
            }
            for ( int i = 0; i < list.size(); i++ ) {
                TaskListener listener = ( TaskListener ) list.elementAt( i );
                listener.taskStopWorking( e );
            }
        }
    }
}
