////////////////////////////////////////////////////////////////////////////////////  
//  
//  Copyright (C) 1996 L.  Patocchi & W.Gander
//
//  This program is free  software;  you can  redistribute  it and/or  modify it
//  under the terms of the GNU General  Public  License as published by the Free
//  Software  Foundation;  either  version 2 of the License, or (at your option)
//  any later version.
//
//  This program is  distributed in the hope that it will be useful, but WITHOUT
//  ANY  WARRANTY;  without  even the implied  warranty  of  MERCHANTABILITY  or
//  FITNESS FOR A PARTICULAR  PURPOSE.  See the GNU General  Public  License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software  Foundation, Inc., 675 Mass
//  Ave, Cambridge, MA 02139, USA.
//
//  Contacts:
//  
//    Project Supervisor
//      W.Hett       hew@info.isbiel.ch
//  
//    Authors
//      W.Gander     gandw@info.isbiel.ch
//      L.Patocchi   patol@info.isbiel.ch
//
//  Documentation can be found at:
//
//      http://www.isbiel.ch/Projects/janet/index.html
//
////////////////////////////////////////////////////////////////////////////////////
//
// File : BPN.java
//
//////////////////////////////////////////////////////////////////
//
//						BPN base class
//
//////////////////////////////////////////////////////////////////
//
//	Author:  Patocchi L.
//	Date:    03.09.1996
//	Project: jaNet
//
//	BPN class is the class that end-user will utilise to
//	integrate a backpropagation jaNet network in his application.
//	
//	- a BPN network is composed by:
//		*  1 input layer
//		*  0 or more hidden layers
//		*  1 output layer
//	
//////////////////////////////////////////////////////////////////
//
//	date		who					what
//  03.09.1996	Patocchi L.			creation

package jaNet.backprop;

import java.util.*;
import java.io.*;

public class BPN {

	private static final String		version = "1.0beta1";

//////////////////////////////////////////////////////////////////
//
//	Protected variables
//
//////////////////////////////////////////////////////////////////
	protected BPNLayer				allLayers[]		= null;
	protected BPNWeightPack			allWeights[]	= null;
	protected inputLayer			inputL;
	protected outputLayer			outputL;
	
	// neural nets functional variables;
	protected 			double		learningRate;
	protected 			double		momentum;
	protected 			double		initMin 		= -0.1;
	protected 			double		initMax			=  0.1;
	// neural nets functional constants;
	public static final double      FIRE 			=  0.999;
	public static final double      NEUTRAL 		=  0.0;
	public static final double      DOWN 			= -0.999;
 
//////////////////////////////////////////////////////////////////
//
//	Private variables
//
//////////////////////////////////////////////////////////////////
	private static boolean trace;
//////////////////////////////////////////////////////////////////
//
//	Constructor
//
//////////////////////////////////////////////////////////////////
	public BPN(int layerDef[], String layerActivatingFn[]) throws BPNException{
		if(layerDef.length != layerActivatingFn.length + 1)
			throw new BPNException("BPN: Error in BPN, arguments have different size.");
		// Backpropagation structure type creation
		buildStructure(layerDef);
		// setup layers activating functions
		setActivationFn(layerActivatingFn);
		// rendomize network weights
		randomize(0.5);
	}
	public BPN(String path, String name) throws BPNException{
		loadNeuralNetwork(path, name);
	}
	
//////////////////////////////////////////////////////////////////
//
//	this build just the structure of desired network. 
//
//		int layerDef[]		specify the layer sizes and must
//							contain at least 2 values
//
//////////////////////////////////////////////////////////////////
	private void buildStructure(int layerDef[]) throws BPNException {
		BPNWeightPack weightP = null;
		hiddenLayer hiddenL;
		
		if(layerDef.length < 2)
			throw new BPNException("BPN: Error in buildStructure, You must define at least input and output layers.");
		for(int i=0; i<layerDef.length; i++)
			if(layerDef[i] < 1) throw new BPNException("BPN: Error in buildStructure, negative or null layer length found..");
		
		allWeights = new BPNWeightPack[layerDef.length-1];
		allLayers = new BPNLayer[layerDef.length];
		for(int i=0; i<layerDef.length; i++){
			 
			if(i==0){
				///////////////////////////////////////
				// define input layer
				///////////////////////////////////////
				if(trace) System.out.println("Create input layer with "+layerDef[i]+" units.");
				inputL = new inputLayer(layerDef[i]);
				allLayers[i] = (BPNLayer) inputL;
				// create lower BPNWeightPack
				weightP = new BPNWeightPack(layerDef[i],layerDef[i+1]);
				allWeights[i] = weightP;
				inputL.setLowerWeightPack(weightP);
				// define which kind of layer is this.
				inputL.setLayerKind(BPNLayer.INPUT);
			}else
			if(i==layerDef.length-1){
				///////////////////////////////////////
				// define output layer
				///////////////////////////////////////
				if(trace) System.out.println("Create output layer with "+layerDef[i]+" units.");
				outputL = new outputLayer(layerDef[i]);
				allLayers[i] = (BPNLayer) outputL;
				// attach upper BPNWeightPack
				outputL.setUpperWeightPack(weightP);
				// attach upper Layer
				outputL.setUpperLayer(allLayers[i-1]);
				// attach upper layer to this layer
				allLayers[i-1].setLowerLayer(allLayers[i]);
				// define which kind of layer is this.
				outputL.setLayerKind(BPNLayer.OUTPUT);
			}else{
				///////////////////////////////////////
				// define hidden layer
				///////////////////////////////////////
				if(trace) System.out.println("Create hidden layer with "+layerDef[i]+" units.");
				hiddenL = new hiddenLayer(layerDef[i]);
				allLayers[i] = (BPNLayer) hiddenL;
				// attach upper BPNWeightPack
				hiddenL.setUpperWeightPack(weightP);
				// create lower BPNWeightPack
				weightP = new BPNWeightPack(layerDef[i],layerDef[i+1]);
				allWeights[i] = weightP;
				// attach lower BPNWeightPack
				hiddenL.setLowerWeightPack(weightP);
				// attach upper layer
				hiddenL.setUpperLayer(allLayers[i-1]);
				// attach upper layer to this layer
				allLayers[i-1].setLowerLayer(allLayers[i]);
				// define which kind of layer is this.
				hiddenL.setLayerKind(BPNLayer.HIDDEN);
			}
		}
		
	}
	private void buildStructure(RandomAccessFile file) throws BPNException {
		BPNWeightPack weightP = null;
		hiddenLayer hiddenL;
		
		try{
			if(file.readUTF().compareTo(getClass().getName()+version) != 0){
				throw new BPNException("BPN: Error in buildStructure, unknown version.");
			}
			learningRate = file.readDouble();
			momentum = file.readDouble();
			initMin = file.readDouble();
			initMax = file.readDouble();
			
			int nlayers = file.readInt();
			allLayers = new BPNLayer[nlayers];
			allWeights = new BPNWeightPack[nlayers-1];
			
			for(int i=0; i<nlayers; i++){
				if(i==0){
					///////////////////////////////////////
					// define input layer
					///////////////////////////////////////
					inputL = inputLayer.readFromFile(file);
					allLayers[i] = (BPNLayer) inputL;
					// create lower BPNWeightPack
					weightP = new BPNWeightPack(file);
					allWeights[i] = weightP;
					inputL.setLowerWeightPack(weightP);
					// define which kind of layer is this.
					inputL.setLayerKind(BPNLayer.INPUT);
				}else
				if(i==nlayers-1){
					///////////////////////////////////////
					// define output layer
					///////////////////////////////////////
					outputL = outputLayer.readFromFile(file);
					allLayers[i] = (BPNLayer) outputL;
					// attach upper BPNWeightPack
					outputL.setUpperWeightPack(weightP);
					// attach upper Layer
					outputL.setUpperLayer(allLayers[i-1]);
					// attach upper layer to this layer
					allLayers[i-1].setLowerLayer(allLayers[i]);
					// define which kind of layer is this.
					outputL.setLayerKind(BPNLayer.OUTPUT);
					// setup neural net parameters
					outputL.setLearningRate(learningRate);
					outputL.setMomentum(momentum);
				}else{
					///////////////////////////////////////
					// define hidden layer
					///////////////////////////////////////
					hiddenL = hiddenLayer.readFromFile(file);
					allLayers[i] = (BPNLayer) hiddenL;
					// attach upper BPNWeightPack
					hiddenL.setUpperWeightPack(weightP);
					// create lower BPNWeightPack
					weightP = new BPNWeightPack(file);
					allWeights[i] = weightP;
					// attach lower BPNWeightPack
					hiddenL.setLowerWeightPack(weightP);
					// attach upper layer
					hiddenL.setUpperLayer(allLayers[i-1]);
					// attach upper layer to this layer
					allLayers[i-1].setLowerLayer(allLayers[i]);
					// define which kind of layer is this.
					hiddenL.setLayerKind(BPNLayer.HIDDEN);
					// setup neural net parameters
					hiddenL.setLearningRate(learningRate);
					hiddenL.setMomentum(momentum);					
				}

				
			}

			
		}catch(IOException ioe){
			throw new BPNException("BPN: Error in buildStructure,("+ioe+").");
		}

	}
//////////////////////////////////////////////////////////////////
//
//	set-up the activating functions for layers
//
//////////////////////////////////////////////////////////////////
	private void setActivationFn(String layerActivatingFn[]) throws BPNException{
		if(allLayers == null)
			throw new BPNException("BPN: Error in setActivationFn, no network structure.");
		for(int i=1; i<allLayers.length; i++)
			setActivationFn(layerActivatingFn[i-1], i);
	}
	public void setActivationFn(String layerActivatingFn, int pos) throws BPNException{
		if(allLayers == null)
			throw new BPNException("BPN: Error in setActivationFn, no network structure.");
		if(pos<1 || pos >= allLayers.length)
			throw new BPNException("BPN: Error in setActivationFn, invalid position.");
			
		allLayers[pos].setActivationFnClass(layerActivatingFn);
	}
//////////////////////////////////////////////////////////////////
//
//	initialisers
//
//////////////////////////////////////////////////////////////////
	public void randomize() throws BPNException{
		randomize(initMin, initMax);
	}
	public void randomize(double value) throws BPNException{
		if(value<0.0) value *= (-1.0);
		randomize(value*(-1.0),value);
	}
	public void randomize(double min, double max) throws BPNException{
		if(allWeights == null)
			throw new BPNException("BPN: Error in randomize, no network structure.");
		for(int i=0; i<allWeights.length; i++)
			allWeights[i].randomize(min, max);
	}
//////////////////////////////////////////////////////////////////
//
//	Backpropagation neural network functionalities
//
//////////////////////////////////////////////////////////////////
	public void propagate() throws BPNException{
		if(inputL == null)
			throw new BPNException("BPN: Error in propagate, no network structure.");
		inputL.propagate();
	}
	public void propagate(double vector[]) throws BPNException{
		if(inputL == null)
			throw new BPNException("BPN: Error in propagate, no network structure.");
		inputL.propagate(vector);
	}
	public double[] getOutputVector() throws BPNException{
		if(outputL == null)
			throw new BPNException("BPN: Error in getOutputVector, no network structure.");
		return outputL.getVector();
	}
	public void learn(double input[], double target[]) throws BPNException{
		outputL.learn(input, target);
	}
//////////////////////////////////////////////////////////////////
//
//	SET's
//
//////////////////////////////////////////////////////////////////
	public void setLearningRate(double val){
		learningRate = val;
		for(int i=1; i<allLayers.length; i++)
			allLayers[i].setLearningRate(val);
	}
	public void setMomentum(double val){
		momentum = val;
		for(int i=1; i<allLayers.length; i++)
			allLayers[i].setMomentum(val);
	}
	public void setInitMin(double val){
		initMin = val;
	}
	public void setInitMax(double val){
		initMax = val;
	}
	public void setBias(int layer, double value) throws BPNException{
		if(layer<0 || layer>allLayers.length-1)
			throw new BPNException("BPN: Error in setBias, layer index is out of bounds");
		allLayers[layer].setBias(value);
	}
	public void setActivationFnClass(int layer, String ActFn) throws BPNException{
		if(layer<0 || layer>allLayers.length-1)
			throw new BPNException("BPN: Error in setActivationFnClass, layer index is out of bounds");
		allLayers[layer].setActivationFnClass(ActFn);
	}
	public void setWeightCopy(){
		for(int i=0; i<allWeights.length; i++){
			allWeights[i].setCopy();
		}
	}
	public void restoreWeightCopy(){
		for(int i=0; i<allWeights.length; i++) allWeights[i].restoreCopy();
	}
//////////////////////////////////////////////////////////////////
//
//	GET's
//
//////////////////////////////////////////////////////////////////
	public double getLearningRate(){
		return learningRate;
	}
	public double getMomentum(){
		return momentum;
	}
	public double getError(){
		return outputL.getError();
	}
	public double getError(double vector[], double target[]) throws BPNException{
		return outputL.getError(vector, target);
	}
	public double getInitMin(){
		return initMin;
	}
	public double getInitMax(){
		return initMax;
	}
	public BPNdescriptor getBPNdescriptor(){
		return new BPNdescriptor(allLayers);
	}
	//public BPNWeightPack[] getAllWeights(){
	//	return allWeights;
	//}
//////////////////////////////////////////////////////////////////
//
//	I/O functionalities
//
//////////////////////////////////////////////////////////////////
	public void saveNeuralNetwork(String path, String name) throws BPNException{
		if(name == null || path == null)
			throw new BPNException("BPN: Error in saveNeuralNetwork, path or name are null.");
		try{
			File rawfile = new File(path, name);
			try{
				RandomAccessFile file = new RandomAccessFile(rawfile,"rw");
				saveNeuralNetwork(file);
				file.close();
			}catch(IllegalArgumentException iae){
				throw new BPNException("BPN: Error in saveNeuralNetwork,"+iae);
			}
		}	
		catch(IOException ioe){
			throw new BPNException("BPN: Error in saveNeuralNetwork,"+ioe);
		}
		catch(SecurityException se){
			throw new BPNException("BPN: Error in saveNeuralNetwork,"+se);
		}
	}
	public void saveNeuralNetwork(RandomAccessFile file) throws IOException,BPNException{
		file.writeUTF(""+getClass().getName()+version);
		// write network basic data
		file.writeDouble(learningRate);
		file.writeDouble(momentum);
		file.writeDouble(initMin);
		file.writeDouble(initMax);
		// write the number of layers we own
		file.writeInt(allLayers.length);
		// for each layer write ....
		for(int i=0;i<allLayers.length;i++){
			// write himself on file
			allLayers[i].writeToFile(file);			
			// the lower weight pack
			if(i != allLayers.length-1)
				allLayers[i].getLowerWeightPack().writeToFile(file);
		}
	}
	public void loadNeuralNetwork(String path, String name) throws BPNException{
		if(name == null || path == null)
			throw new BPNException("BPN: Error in loadNeuralNetwork, path or name are null.");
		try{
			File rawfile = new File(path, name);
			try{
				RandomAccessFile file = new RandomAccessFile(rawfile,"r");
				loadNeuralNetwork(file);
				file.close();
			}catch(IllegalArgumentException iae){
				throw new BPNException("BPN: Error in loadNeuralNetwork,"+iae);
			}
		}	
		catch(IOException ioe){
			throw new BPNException("BPN: Error in loadNeuralNetwork,"+ioe);
		}
		catch(SecurityException se){
			throw new BPNException("BPN: Error in loadNeuralNetwork,"+se);
		}
	}
	public void loadNeuralNetwork(RandomAccessFile file) throws BPNException{
		buildStructure(file);
	}

}
